Package: rtrees
Title: Deriving Phylogenies from Synthesis Trees
Version: 2.0.2
Authors@R: 
    person(given = "Daijiang",
           family = "Li",
           role = c("aut", "cre"),
           email = "daijianglee@gmail.com",
           comment = c(ORCID = "0000-0002-0925-3421"))
Description: Provides tools to derive species-level phylogenies from large
    synthesis mega-trees for a wide range of taxonomic groups, including
    plants, birds, mammals, amphibians, reptiles, fish, bees, butterflies,
    and sharks. When a queried species is absent from the mega-tree, it is
    grafted onto the tree using one of two placement strategies: attachment
    at the basal node of the most closely related genus or family
    ('at_basal_node'), or random attachment below that basal node with
    probability proportional to branch length ('random_below_basal'). 
    See Li (2023) <doi:10.1111/ecog.06643> for details. Multiple species from a 
    genus not represented in the mega-tree are placed as a polytomy to preserve 
    clade coherence. The package interfaces with the 'megatrees' data package,
    which bundles or downloads on demand curated mega-trees. 
    Users can also provide their own mega-trees.
License: GPL-3
Encoding: UTF-8
LazyData: true
Depends: R (>= 3.5.0)
Imports: ape, tidytree, dplyr, tibble, utils, castor, furrr, future,
        megatrees (>= 0.1.3), fastmatch, data.table, Rcpp
LinkingTo: Rcpp
Suggests: knitr, rmarkdown, testthat, piggyback, R.rsp, ggplot2
URL: https://daijiang.github.io/rtrees/
VignetteBuilder: knitr, R.rsp
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-06-05 15:40:11 UTC; dli
Author: Daijiang Li [aut, cre] (ORCID: <https://orcid.org/0000-0002-0925-3421>)
Maintainer: Daijiang Li <daijianglee@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-11 11:40:07 UTC

# Generated by roxygen2: do not edit by hand

export(add_root_info)
export(bind_tip)
export(bind_tip_df)
export(get_graft_status)
export(get_one_tree)
export(get_tree)
export(rm_stars)
export(sp_list_df)
import(data.table)
importFrom(Rcpp,evalCpp)
importFrom(fastmatch,"%fin%")
importFrom(stats,setNames)
useDynLib(rtrees, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rtrees

The goal of `rtrees` is to remove the barriers to derive phylogenies for
a list of species from mega-trees. Basically, `Phylomatic` and more.

**`rtrees` just got an important update to speed it up! Please try to
install the latest version. If you see any problem, please file a GitHub
issue. Thanks!**

# Installation

``` r
options(repos = c(
  rtrees = 'https://daijiang.r-universe.dev',
  CRAN = 'https://cloud.r-project.org'))
install.packages("rtrees")
```

The above code will also install one dependency
[{megatrees}](https://github.com/daijiang/megatrees), an R data package
that hosts the megatrees used here.

## Shinny app

There is also a [Shinny app](https://djli.shinyapps.io/rtrees_shiny/) to
get phylogenies without using R. However, given that I can only afford
the free plan, the number of species to be inserted to the megatrees
here will be limited to 1,000.

# Mega-trees

Currently, `rtrees` provides mega-trees for the following taxonomic
groups: amphibian, bird, fish, mammal, plant, reptile, and shark_ray.
The mega-trees (or subset of 100 posterior phylogenies; 50 for fish to
reduce the package size) are stored as GitHub release assets in the data
package [{megatrees}](https://github.com/daijiang/megatrees) and
downloaded on demand via getter functions.

| Taxon | \# of species | \# of trees | How to load (in `{megatrees}`) | Reference |
|----|----|----|----|----|
| Amphibian | 7238 | 100 | `get_tree_amphibian_n100()` | Jetz and Pyron 2018 |
| Bee | 4651 | 1 | `tree_bee` | Henríquez-Piskulich et al. 2023 |
|  | 4651 | 100 | `get_tree_bee_n100()` | Henríquez-Piskulich et al. 2023 |
| Butterfly | 2244 | 1 | `tree_butterfly` | Kawahara et al. 2023 |
| Bird | 9993 | 100 | `get_tree_bird_n100()` | Jetz et al. 2012 |
|  | 11017 | 1 | `tree_bird_McTavish` | McTavish et al. 2025 |
| Fish | 11638 | 1 | `tree_fish_12k` | Rabosky et al. 2018 |
|  | 31516 | 50 | `get_tree_fish_32k_n50()` | Rabosky et al. 2018 |
| Mammal | 5831 | 100 | `get_tree_mammal_n100_phylacine()` | Faurby et al. 2018 |
|  | 5911 | 100 | `get_tree_mammal_n100_vertlife()` | Upham et al. 2019 |
| Plant | 74531 | 1 | `tree_plant_otl` (default) | Smith and Brown 2018 |
|  | 123182 | 1 | `tree_plant_Carruthers` | Carruthers et al. 2026 |
|  | 123182 | 100 | `get_tree_plant_n100_Carruthers()` | Carruthers et al. 2026 |
| Reptile (Squamate) | 9755 | 100 | `get_tree_reptile_n100()` | Tonini et al. 2016 |
| Shark, Ray, and Chimaera | 1192 | 100 | `get_tree_shark_ray_n100()` | Stein et al. 2018 |

# Usage examples

``` r
library(rtrees)
library(ape)
```

## Prepare species data frame (optional)

The species lists which we want to have a phylogeny should be provided
as a data frame with at least 3 columns: `family`, `genus`, and
`species`. Their order does not matter. Here is an example for fish.

``` r
test_fish_list = tibble::tibble(
  species = c("Serrasalmus_geryi", "Careproctus_reinhardti", "Gobiomorphus_coxii", 
              "Periophthalmus_barbarus", "Prognichthys_glaphyrae", "Barathronus_bicolor", 
              "Knipowitschia_croatica", "Rhamphochromis_lucius", "Neolissochilus_tweediei", 
              "Haplochromis_nyanzae", "Astronesthes_micropogon", "Sanopus_reticulatus"),
  genus = c("Serrasalmus", "Careproctus", "Gobiomorphus", "Periophthalmus",
            "Prognichthys", "Barathronus", "Knipowitschia", "Rhamphochromis", 
            "Neolissochilus", "Haplochromis", "Astronesthes", "Sanopus"),
  family = c("Serrasalmidae", "Liparidae", "Eleotridae", "Gobiidae", 
             "Exocoetidae", "Aphyonidae", "Gobiidae", "Cichlidae", 
             "Cyprinidae", "Cichlidae", "Stomiidae", "Batrachoididae")
)
```

``` r
test_fish_list
#> # A tibble: 12 × 3
#>    species                 genus          family        
#>    <chr>                   <chr>          <chr>         
#>  1 Serrasalmus_geryi       Serrasalmus    Serrasalmidae 
#>  2 Careproctus_reinhardti  Careproctus    Liparidae     
#>  3 Gobiomorphus_coxii      Gobiomorphus   Eleotridae    
#>  4 Periophthalmus_barbarus Periophthalmus Gobiidae      
#>  5 Prognichthys_glaphyrae  Prognichthys   Exocoetidae   
#>  6 Barathronus_bicolor     Barathronus    Aphyonidae    
#>  7 Knipowitschia_croatica  Knipowitschia  Gobiidae      
#>  8 Rhamphochromis_lucius   Rhamphochromis Cichlidae     
#>  9 Neolissochilus_tweediei Neolissochilus Cyprinidae    
#> 10 Haplochromis_nyanzae    Haplochromis   Cichlidae     
#> 11 Astronesthes_micropogon Astronesthes   Stomiidae     
#> 12 Sanopus_reticulatus     Sanopus        Batrachoididae
```

For plant, fish, bird, and mammal, it is possible to prepare `sp_list`
with function `sp_list_df()`; though you should check the results. It is
possible to just provide the speices as a character vector for these
taxon groups (or all genus in your species list are presented in the
phylogeny).

``` r
sp_list_df(sp_list = c("Periophthalmus_barbarus", "Barathronus_bicolor"),
           taxon = "fish")
#> # A tibble: 2 × 3
#>   species                 genus          family    
#>   <chr>                   <chr>          <chr>     
#> 1 Periophthalmus_barbarus Periophthalmus Gobiidae  
#> 2 Barathronus_bicolor     Barathronus    Aphyonidae
```

## Get phylogeny from one megatree

Once we have the species list ready, we can then derive a phylogeny from
`tree_fish`.

``` r
test_tree = get_tree(sp_list = test_fish_list,
                     taxon = "fish",
                     scenario = "at_basal_node",
                     show_grafted = TRUE)
#> 
#> 6 species added at genus level (*)
#> 1 species have no co-family species in the mega-tree, skipped
#> (if you know their family, prepare and edit species list with `rtrees::sp_list_df()` may help): 
#> Barathronus_bicolor
plot(test_tree, no.margin = T)
```

<img src="man/figures/README-unnamed-chunk-6-1.png" alt="" width="100%" />

No matter whether `show_grafted` is ‘TRUE’ or ‘FALSE’, the grafting
information was saved along with the phylogeny and can be extracted with
the following code:

``` r
# or use rtrees::get_graft_status(test_tree)
test_tree$graft_status
#> # A tibble: 12 × 3
#>    tip_label                species                 status                      
#>    <chr>                    <chr>                   <chr>                       
#>  1 Prognichthys_glaphyrae   Prognichthys_glaphyrae  exisiting species in the me…
#>  2 Periophthalmus_barbarus  Periophthalmus_barbarus exisiting species in the me…
#>  3 Gobiomorphus_coxii       Gobiomorphus_coxii      exisiting species in the me…
#>  4 Careproctus_reinhardti   Careproctus_reinhardti  exisiting species in the me…
#>  5 Sanopus_reticulatus*     Sanopus_reticulatus     grafted at genus level      
#>  6 Serrasalmus_geryi        Serrasalmus_geryi       exisiting species in the me…
#>  7 Knipowitschia_croatica*  Knipowitschia_croatica  grafted at genus level      
#>  8 Rhamphochromis_lucius*   Rhamphochromis_lucius   grafted at genus level      
#>  9 Neolissochilus_tweediei* Neolissochilus_tweediei grafted at genus level      
#> 10 Haplochromis_nyanzae*    Haplochromis_nyanzae    grafted at genus level      
#> 11 Astronesthes_micropogon* Astronesthes_micropogon grafted at genus level      
#> 12 <NA>                     Barathronus_bicolor     skipped as no co-family in …
```

## Get a set of phylogenies from multiple posterior megatrees

The function `rtrees::get_tree()` can also work with a set of posterior
megatrees with the option to use parallel computing for the whole
process. The default number of cores to be used will be the available
number of cores minus 2 (so that users can still perform other tasks on
their computers at the same time). The output will be a set of generated
phylogenies with class `multiPhylo`; the number of derived phylogenies
will be the same as the input megatrees. For this scenario, we can use
exactly the same code described above. For example, if we want to use
the 100 randomly selected posterior phylogenies that have ~32k fish
species, we can add the `fish_tree = "all-taxon`.

``` r
test_tree2 = get_tree(sp_list = test_fish_list,
                     taxon = "fish",
                     scenario = "at_basal_node",
                     show_grafted = TRUE,
                     fish_tree = "all-taxon")
#> Wow, all species are already in the mega-tree!
test_tree2
#> 50 phylogenetic trees
plot(ladderize(test_tree2[[1]]), no.margin = T)
```

<img src="man/figures/README-unnamed-chunk-8-1.png" alt="" width="100%" />

## Add tips to user provided trees

It is also possible to derive a phylogeny based on a user provided tree.
This can be particularly useful if we have things like `Acer_spp` that
we want to insert to the basal node of the genus.

``` r
test_tree = ape::read.tree(text = "(((((((((Potentilla_intermedia:1.156097,Potentilla_gracilis:1.156097):9.169741,Potentilla_bipinnatifida:10.325839):10.466967,Potentilla_sterilis:20.792806):11.333216,Fragaria_virginiana:32.126022):6.026567,(((((Rosa_setigera:3.434279,Rosa_arkansana:3.434279):1.991106,Rosa_spinosissima:5.425386):0.993924,Rosa_acicularis:6.419309):6.730804,(Rosa_micrantha:0.188519,Rosa_canina:0.188519):12.961594):23.677485,(Agrimonia_gryposepala:4.730863,Agrimonia_parviflora:4.730863):32.096735):1.324992):6.47604,(Geum_urbanum:11.153303,Geum_rivale:11.153303):33.475326):1.449852,((((((((Rubus_semisetosus:3.10529,Rubus_glandicaulis:3.10529):0.957584,Rubus_steelei:4.062873):0.398394,Rubus_cuneifolius:4.461267):2.271628,Rubus_vagus:6.732895):0.023439,Rubus_superioris:6.756333):0.020104,(Rubus_multifer:4.287114,Rubus_elegantulus:4.287115):2.489323):0.025475,((Rubus_laciniatus:0.125186,Rubus_bifrons:0.125186):0.11903,Rubus_praecox:0.244216):6.557697):11.571849,(Rubus_parviflorus:14.85915,Rubus_odoratus:14.85915):3.514612):27.704718):16.915234,Filipendula_rubra:62.993714):13.354116,((((((((((Crataegus_dissona:0.081863,Crataegus_florifera:0.081863):0.49838,Crataegus_fulleriana:0.580243):1.434261,Crataegus_crus-galli:2.014503):0.177737,Crataegus_chrysocarpa:2.192239):0.051159,Crataegus_uniflora:2.243399):0.438351,Crataegus_persimilis:2.68175):4.697259,((Malus_toringo:6.472694,Malus_ioensis:6.472693):0.772012,Chaenomeles_speciosa:7.244705):0.134305):1.72489,Pyrus_communis:9.1039):0.313393,Amelanchier_spicata:9.417293):40.039557,((((((Prunus_pumila:6.189944,Prunus_nigra:6.189945):0.37487,Prunus_pensylvanica:6.564815):4.856241,(Prunus_padus:4.816999,Prunus_serotina:4.817):6.604057):29.162527,Sorbaria_sorbifolia:40.583584):2.924475,Rhodotypos_scandens:43.50806):2.718224,(Spiraea_X_bumalda:25.921582,Aruncus_dioicus:25.921583):20.304701):3.230567):26.89098);")

test_tree_sp = c("Rosa_sp", "Rubus_sp", "Amelanchier_sp", "Prunus_sp", "Sorbus_sp", "Potentilla_sp",
  "Potentilla_intermedia", "Potentilla_gracilis", "Potentilla_bipinnatifida", "Potentilla_sterilis", "Fragaria_virginiana", "Rosa_setigera", "Rosa_arkansana", "Rosa_spinosissima", "Rosa_acicularis", "Rosa_micrantha", "Rosa_canina", "Agrimonia_gryposepala", "Agrimonia_parviflora", "Geum_urbanum", "Geum_rivale", "Rubus_semisetosus", "Rubus_glandicaulis", "Rubus_steelei", "Rubus_cuneifolius", "Rubus_vagus", "Rubus_superioris", "Rubus_multifer", "Rubus_elegantulus", "Rubus_laciniatus", "Rubus_bifrons", "Rubus_praecox", "Rubus_parviflorus", "Rubus_odoratus", "Filipendula_rubra", "Crataegus_dissona", "Crataegus_florifera", "Crataegus_fulleriana", "Crataegus_crus-galli", "Crataegus_chrysocarpa", "Crataegus_uniflora", "Crataegus_persimilis", "Malus_toringo", "Malus_ioensis", "Chaenomeles_speciosa", "Pyrus_communis", "Amelanchier_spicata", "Prunus_pumila", "Prunus_nigra", "Prunus_pensylvanica", "Prunus_padus", "Prunus_serotina", "Sorbaria_sorbifolia", "Rhodotypos_scandens", "Spiraea_X_bumalda", "Aruncus_dioicus")

plot(get_tree(sp_list = test_tree_sp, tree = test_tree, taxon = "plant",
              show_grafted = T, tree_by_user = T), type = "fan")
#> Not all genus can be found in the phylogeny.
#> Warning: For user provided phylogeny, without a classification for all genus of species in the phylogeny,
#>               it is unlikely to find the most recent ancestor for genus and family; here we proceed the phylogeny
#>               by adding root information for genus and family that can be found in the phylogeny or species list but
#>               we recommend to prepare the phylogeny using `add_root_info()` with a classification
#>               data frame with all tips first.
#> 
#> 5 species added at genus level (*)
#> 1 species added at family level (**)
```

<img src="man/figures/README-unnamed-chunk-9-1.png" alt="" width="100%" />

**Pay attention to the warning message there**. Because not all genera
in the species list are presented in the megatree, it is impossible to
find the most recent ancestor for genus and family that are missing from
the megatree. Therefore, **for user-provided phylogenies, here is the
recommended steps**:

1.  Get the classification information (a data frame with two columns:
    genus and family) for all genera presented in the species list and
    the megatree.
2.  Prepare the species list as a data frame with three columns:
    species, genus, and family using the classification information.
3.  Process the megatree using function
    `processed_tree = rtrees::add_root_infor(tree, classification)`.
4.  Pass the processed megatree to
    `rtrees::get_tree(sp_list_data_frame, processed_tree)` to derive the
    phylogeny for your species list.

## Bind missing species to specified places

It is also possible to specify a particular species to bind with by
specifying columns `close_sp` and/or `close_genus`.

``` r
test_tree_sp_df = sp_list_df(test_tree_sp, "plant")
test_tree_sp_df$close_sp = NA
test_tree_sp_df$close_sp[1] = "Rosa_acicularis" # bind Rosa_sp to here
plot(get_tree(sp_list = test_tree_sp_df, tree = test_tree, taxon = "plant",
              show_grafted = T, tree_by_user = T), type = "fan")
#> Not all genus can be found in the phylogeny.
#> Warning: For user provided phylogeny, without a classification for all genus of species in the phylogeny,
#>               it is unlikely to find the most recent ancestor for genus and family; here we proceed the phylogeny
#>               by adding root information for genus and family that can be found in the phylogeny or species list but
#>               we recommend to prepare the phylogeny using `add_root_info()` with a classification
#>               data frame with all tips first.
#> 
#> 5 species added at genus level (*)
#> 1 species added at family level (**)
```

<img src="man/figures/README-unnamed-chunk-10-1.png" alt="" width="100%" />

# Some notes

- If `tree` is specified, then `taxon` can be ignored if all genus in
  the species list are presented in the phylogeny.
- If a species does not have a co-family species in the mega-tree, it
  will be skipped.
- If `show_grafted = TRUE`, species that are grafted will have one or
  two `*` at the end of their names.
  - If it is grafted at the genus level, one `*`.
  - If it is grafted at the family level, two `*`s (i.e., `**`).
- The default scenario is `at_basal_node`, which will graft species at
  the genus/family basal node.
  - `random_below_basal` will randomly select a downstream node to
    attach the new tip.
  - If only one species in the mega-tree that is in the same
    genus/family of the new tip, then the new tip will be grafted at the
    middle of this species’ branch for all scenarios.
- The `tree` can be a user provided tree, if so, set
  `tree_by_user = TRUE`.
- See `?get_tree` for more details.

# Contribution

Feel free to test it. Contributions and suggestions are welcome. You can
open an issue or send a pull request.

# Citation

Li Daijiang. (2023). [rtrees: an R package to assemble phylogenetic
trees from
megatrees](https://nsojournals.onlinelibrary.wiley.com/doi/10.1111/ecog.06643).
Ecography. e06643

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----out.width='90%', fig.width=7, fig.height=5-------------------------------
library(ggplot2)
speed_out = readRDS(system.file("extdata", "rtrees_speed_out.rds", package = "rtrees"))
speed_out = dplyr::mutate(speed_out, n_sp_missing_k = n_sp_missing / 1000, time_s = time/1e9)
# speed_lm = dplyr::group_by(speed_out, expr) |> 
#   dplyr::do(broom::tidy(lm(time_s ~ n_sp_missing_k, data = .))) |> 
#   dplyr::ungroup()
ggplot(speed_out, aes(x = n_sp_missing_k, y = time_s, color = expr, group = expr)) +
  geom_point() +
  geom_smooth(method = "lm") +
  labs(x = "Number of binding species (Thousand)",
       y = "Time (Second)", color = NULL, group = NULL) +
  scale_x_continuous(breaks = c(0.05, seq(0.5, 5, 0.5))) +
  scale_y_continuous(breaks = c(0, 10, seq(25, 200, 25))) +
  geom_text(x = 4, y = 30, label = "rtrees: y = 0.35+ 0.95x", inherit.aes = FALSE, color = "#F8766D") +
  geom_text(x = 1.8, y = 150, label = "V.PhyloMaker2: y = 12.3 + 36.8x", inherit.aes = FALSE, color = "#00BFC4") +
  theme(legend.position = "none") 

# Packages required: phylocomr, microbenchmark, tidyverse, V.PhyloMaker2
# Install them manually before running this script if needed.
library(microbenchmark)
library(tidyverse)
library(V.PhyloMaker2)

d = read_delim("https://raw.githubusercontent.com/daijiang/rtrees_ms/refs/heads/main/Data/plants_phylomatic_names.txt", "/", col_names = F) %>%
  set_names(c("family", "genus", "species")) %>%
  mutate(genus = str_extract(genus, "^[^_]+"),
         species = rtrees:::cap_first_letter(species),
         genus = rtrees:::cap_first_letter(genus),
         family = rtrees:::cap_first_letter(family)) %>%
  select(species) %>% distinct()
n_distinct(d$species)

sp_missing = setdiff(d$species, megatrees::tree_plant_otl$tip.label)
sp_in = intersect(d$species, megatrees::tree_plant_otl$tip.label)

#' Function to test speed
#' 
#' @param n_sp_missing The number of missing species to be binded
#' @param n_sp_in The number of species that were already in the megatree
#' @param n_times How many times to compare? Default of 5 should be enough.
#' @return A data frame with the test methods, and their time used.
#' 
speed_test = function(n_sp_missing, n_sp_in = 500, n_times = 5, ...){
  test_sp = c(sample(sp_in, n_sp_in), sample(sp_missing, n_sp_missing))
  test_sp_df = rtrees::sp_list_df(sp_list = test_sp, taxon = "plant")
  test_sp_df = filter(test_sp_df, !is.na(family))
  test_sp_df2 = mutate(test_sp_df, phylom_sp = paste(family, genus, species, sep = "/"),
                       phylom_sp = casefold(phylom_sp))
  # compare speed
  xx2 = microbenchmark::microbenchmark(
    # Phylomatic = phylocomr::ph_phylomatic(taxa = test_sp_df2$phylom_sp, phylo = megatrees::tree_plant_otl),
    rtrees = get_tree(sp_list = test_sp_df, taxon = "plant"),
    # V.PhyloMaker = V.PhyloMaker::phylo.maker(sp.list = test_sp_df, scenarios = "S1"),
    V.PhyloMaker2 = V.PhyloMaker2::phylo.maker(sp.list = test_sp_df, scenarios = "S1"),
    times = n_times, ...
  )
  
  xx2 = as.data.frame(xx2) %>% 
    mutate(n_sp_in = n_sp_in, n_sp_missing = n_sp_missing)
  
  xx2
}

x = get_tree(sp_list = test_sp_df, taxon = "plant")
y = V.PhyloMaker2::phylo.maker(sp.list = test_sp_df, scenarios = "S1")
plot(x, type = "fan", show.tip.label = F, main = "rtrees")
plot(ape::ladderize(y[[1]]), type = "fan", show.tip.label = F, main = "v.p")

# all tests have 500 species already in the megatree, then
# start with 50 missing species, and 100, 200, 500, 1000, 3000, 5000
speed_out2 = purrr::map_dfr(c(50, 100, 200, 500, 1000, 3000, 5000), speed_test)
speed_out3 = purrr::map_dfr(c(3000, 5000), speed_test, n_times = 3)
speed_out = bind_rows(speed_out2, speed_out3)
# for some reason, phylocomr::ph_phylomatic does not work (no mssin

speed_out = dplyr::mutate(speed_out, n_sp_missing_k = n_sp_missing / 1000, time_s = time/1e9)
saveRDS(speed_out, "inst/extdata/rtrees_speed_out.rds")

speed_lm = dplyr::group_by(speed_out, expr) |>
  dplyr::do(broom::tidy(lm(time_s ~ n_sp_missing_k, data = .))) |>
  dplyr::ungroup()

ggplot(speed_out, aes(x = n_sp_missing_k, y = time_s, color = expr, group = expr)) +
  geom_point() +
  geom_smooth(method = "lm") +
  labs(x = "Number of binding species (Thousand)",
       y = "Time (Second)", color = NULL, group = NULL) +
  scale_x_continuous(breaks = c(0.05, seq(0.5, 5, 0.5))) +
  scale_y_continuous(breaks = c(0, 10, seq(25, 200, 25))) +
  geom_text(x = 4, y = 30, label = "rtrees: y = 0.35+ 0.95x", inherit.aes = FALSE, color = "#F8766D") +
  geom_text(x = 1.8, y = 150, label = "V.PhyloMaker2: y = 12.3 + 36.8x", inherit.aes = FALSE, color = "#00BFC4") +
  cowplot::theme_cowplot() +
  theme(legend.position = "none")  

# profile.R — identify hot spots in get_one_tree / bind_tip
# Run from repo root: Rscript profile.R
suppressPackageStartupMessages({
  library(ape)
  library(devtools)
})
devtools::load_all(".", quiet = TRUE)

load("debug_data.RData")
sp_debug <- rtrees::sp_list_df(true_tree$tip.label, taxon = "plant")
plant_tree <- megatrees::tree_plant_otl
fish_tree  <- megatrees::tree_fish_12k

sp_fish <- tibble::tibble(
  species = c("Serrasalmus_geryi", "Careproctus_reinhardti", "Gobiomorphus_coxii",
               "Periophthalmus_barbarus", "Prognichthys_glaphyrae", "Barathronus_bicolor",
               "Knipowitschia_croatica", "Rhamphochromis_lucius", "Neolissochilus_tweediei",
               "Haplochromis_nyanzae", "Astronesthes_micropogon", "Sanopus_reticulatus"),
  genus  = c("Serrasalmus", "Careproctus", "Gobiomorphus", "Periophthalmus",
              "Prognichthys", "Barathronus", "Knipowitschia", "Rhamphochromis",
              "Neolissochilus", "Haplochromis", "Astronesthes", "Sanopus"),
  family = c("Serrasalmidae", "Liparidae", "Eleotridae", "Gobiidae",
              "Exocoetidae", "Aphyonidae", "Gobiidae", "Cichlidae",
              "Cyprinidae", "Cichlidae", "Stomiidae", "Batrachoididae")
)

# ---------------------------------------------------------------------------
# Profile helper
# ---------------------------------------------------------------------------
run_prof <- function(label, expr, interval = 0.005) {
  f <- tempfile(fileext = ".out")
  Rprof(f, interval = interval, memory.profiling = TRUE)
  suppressMessages(force(expr))
  Rprof(NULL)
  sm <- summaryRprof(f, memory = "both")
  cat("\n===", label, "=== (by.total, top 20)\n")
  top <- head(sm$by.total, 20)
  print(top[, c("total.time", "total.pct", "mem.total")])
  cat("\n--- by.self, top 15 ---\n")
  top2 <- head(sm$by.self, 15)
  print(top2[, c("self.time", "self.pct")])
  invisible(sm)
}

# ---------------------------------------------------------------------------
# 1. Debug data: 130 spp grafted onto 304-tip user tree  (at_basal_node)
# ---------------------------------------------------------------------------
p1 <- run_prof("debug/plant 130sp at_basal_node",
  get_one_tree(sp_debug, tree = reduced_tree, taxon = "plant",
               scenario = "at_basal_node", tree_by_user = TRUE))

# ---------------------------------------------------------------------------
# 2. Fish 12sp on 12k-tip megatree  (at_basal_node)
# ---------------------------------------------------------------------------
p2 <- run_prof("fish 12sp at_basal_node",
  get_one_tree(sp_fish, tree = fish_tree, taxon = "fish",
               scenario = "at_basal_node"))

# ---------------------------------------------------------------------------
# 3. Fish 12sp on 12k-tip megatree  (random_below_basal)
# ---------------------------------------------------------------------------
p3 <- run_prof("fish 12sp random_below_basal", {
  set.seed(1)
  get_one_tree(sp_fish, tree = fish_tree, taxon = "fish",
               scenario = "random_below_basal")
})

# ---------------------------------------------------------------------------
# 4. Plant 12sp on full plant megatree  (at_basal_node) — megatree is ~75k tips
# ---------------------------------------------------------------------------
sp_plant <- rtrees::sp_list_df(
  c("Meliosma_laui", "Acer_cordatum", "Fraxinus_mandshurica", "Ormosia_pinnata",
    "Aglaia_dasyclada", "Claoxylon_indicum",
    "Carya_alba", "Carya_cordiformis", "Stewartia_sinensis",
    "Pterocarya_hupehensis", "Ulmus_davidiana", "Diplospora_dubia"),
  taxon = "plant"
)
p4 <- run_prof("plant 12sp on 75k-tip megatree at_basal_node",
  get_one_tree(sp_plant, tree = plant_tree, taxon = "plant",
               scenario = "at_basal_node"))

cat("\n=== DONE ===\n")

#' Bind a tip to a phylogeny
#'
#' Graft a tip to a phylogeny at location specified.
#'
#' @param tree A phylogeny, with class of "phylo".
#' @param where Location where to insert the tip. It can be either tip label or node label, but must be characters. If the location does not have a name, assign it first.
#' @param tip_label Name of the new tip inserted.
#' @param frac The fraction of branch length, must be between 0 and 1. This only applies when location is a tip or `new_node_above = TRUE`. The distance from the new inserted node to the location (a node or a tip) is the branch length of the location * (1 - frac).
#' @param new_node_above Whether to insert the new node above when the location is a node? Default is `FALSE`, which will attach the new tip to the location node.
#' @param node_label Name of the new node created. This only applies when location is a tip or `new_node_above = TRUE`.
#' @param return_tree Whether to return a phylogeny with class "phylo?" Default is `TRUE`. Otherwise, it will return a data frame.
#' @param tree_tbl A tibble version of the tree, optional.
#' @param node_heights A named numeric vector of node hieghts of the tree, generated by [ape::branching.times()]. It is also optional if `tree` is specified; but required if `tree_tbl` is specified.
#' @param use_castor Whether to use package `castor` to get the phylogeny at a node; it is faster than `tidytree::offspring` to figure out what are the tip offsprings at a node.
#' @param sequential Whether to add the tip with sequential node number in the edge matrix. For example, if we want to bind a tip to a clade and the node number of the tips of this clade is from 101 to 150. We can set the node id of the new tip to 151 and push all the remaining node id to 1 after their current values. This will require us to find out the node ids of all tips that are descents of the node where we want to bind the new tip to, and it can be time costly. Yet I am still not sure whether this is necessary. Normally, the node ids of the `phylo` class are sequential. Therefore, the default value here is `TRUE`. If set to `FALSE`, we can just assign the id of the new tip to Ntip + 1 to save time. In addition, we probably don't need to order the node column of the edge matrix every time.
#' @return Either a phylogeny or a data frame, which can be then converted to a phylogeny later.
#' @examples
#' tr <- ape::read.tree(text = "((A:1,B:1):1,C:2);")
#' tr$node.label <- c("root", "N1")
#' bind_tip(tr, where = "N1", tip_label = "D")
#' tr_tbl <- tidytree::as_tibble(tr)
#' node_hts <- ape::branching.times(tr)
#' bind_tip(tree_tbl = tr_tbl, where = "N1",
#'          tip_label = "D", node_heights = node_hts)
#' @export
#'
bind_tip = function(tree = NULL, where, tip_label,
                    frac = 0.5, new_node_above = FALSE,
                    node_label = NULL, return_tree = TRUE,
                    tree_tbl = NULL, node_heights = NULL, 
                    use_castor = TRUE, sequential = TRUE){
  if(frac > 1 | frac < 0) stop("frac must be between 0 and 1.")
  if(is.null(tree) & is.null(tree_tbl)) stop("No input tree, please assign one.")
  if(!is.null(tree_tbl) & is.null(node_heights)) stop("Please specify node_heights.")
  if(!is.null(tree_tbl) & new_node_above & is.null(node_label)) 
    stop("new_node_above is TRUE but no node_label.")
  if(!is.null(tree_tbl) & !new_node_above & !is.null(node_label)) 
    stop("new_node_above is FALSE but with node_label.")
  if(!is.null(tree)){
    # make sure node labels are assigned and unique
    if(is.null(tree$node.label) || any(duplicated(tree$node.label)))
        tree$node.label = paste0("N", 1:ape::Nnode(tree))
    tree = ape::makeLabel(tree, tips = FALSE, nodes = TRUE)
    tree_tbl = tidytree::as_tibble(tree)
    node_heights = ape::branching.times(tree)
  }
  if(is.null(node_label)) node_label = paste0("node", length(node_heights) + 1L)
  if(!"is_tip" %in% names(tree_tbl)){
    tree_tbl$is_tip = !(tree_tbl$node %fin% tree_tbl$parent)
  }

  tree_tbl_node = tree_tbl[tree_tbl$label == where, ]
  at_root = tree_tbl_node$parent == tree_tbl_node$node
  node_height = node_heights[where]

  if(!tree_tbl_node$is_tip & !at_root & sequential){
    if(use_castor){
      where_offspring = tree_tbl[tree_tbl$label %fin%
                                   castor::get_subtree_at_node(as_tree(tree_tbl), node = where)$subtree$tip.label, ]
      max_offspring = max(where_offspring$node)
    } else {
      where_offspring = tidytree::offspring(tree_tbl, .node = where)
      max_offspring = max(where_offspring$node[where_offspring$is_tip])
    }
  }

  if(at_root & new_node_above & !tree_tbl_node$is_tip)
    message("New tip cannot be inserted above the root; attached instead.")

  # Delegate all arithmetic to C++ — two single-pass loops instead of multiple
  # [.data.table calls, avoiding per-call dispatch overhead on large tables.
  res <- bind_tip_core_cpp(
    parent        = tree_tbl$parent,
    node          = tree_tbl$node,
    branch_length = tree_tbl$branch.length,
    label         = tree_tbl$label,
    is_tip        = tree_tbl$is_tip,
    where         = where,
    tip_label     = tip_label,
    node_label_str = node_label,
    frac          = frac,
    new_node_above = new_node_above,
    node_height   = node_height
  )
  tree_tbl_2 <- tibble::as_tibble(res)
  if(!inherits(tree_tbl_2, "tbl_tree")) class(tree_tbl_2) <- c("tbl_tree", class(tree_tbl_2))

  if(return_tree) return(as_tree(tree_tbl_2))
  tree_tbl_2
}

#' Bind a tip to a phylogeny (data frame version)
#'
#' Graft a tip to a phylogeny at location specified.
#'
#' @param tree A phylogeny, with class of "phylo".
#' @param where Location where to insert the tip. It can be either tip label or node label, but must be characters. If the location does not have a name, assign it first.
#' @param tip_label Name of the new tip inserted.
#' @param frac The fraction of branch length, must be between 0 and 1. This only applies when location is a tip or `new_node_above = TRUE`. The distance from the new inserted node to the location (a node or a tip) is the branch length of the location * (1 - frac).
#' @param new_node_above Whether to insert the new node above when the location is a node? Default is `FALSE`, which will attach the new tip to the location node.
#' @param node_label Name of the new node created. This only applies when location is a tip or `new_node_above = TRUE`.
#' @param return_tree Whether to return a phylogeny with class "phylo?" Default is `TRUE`. Otherwise, it will return a data frame.
#' @param tree_tbl A tibble version of the tree, optional.
#' @param node_heights A named numeric vector of node hieghts of the tree, generated by [ape::branching.times()]. It is also optional if `tree` is specified; but required if `tree_tbl` is specified.
#' @param use_castor Whether to use package `castor` to get the phylogeny at a node; it is faster than `tidytree::offspring` to figure out what are the tip offsprings at a node.
#' @return Either a phylogeny or a data frame, which can be then converted to a phylogeny later.
#' @examples
#' tr <- ape::read.tree(text = "((A:1,B:1):1,C:2);")
#' tr$node.label <- c("root", "N1")
#' bind_tip_df(tr, where = "N1", tip_label = "D")
#' tr_tbl <- tidytree::as_tibble(tr)
#' node_hts <- ape::branching.times(tr)
#' bind_tip_df(tree_tbl = tr_tbl, where = "N1",
#'             tip_label = "D", node_heights = node_hts)
#' @export
#'
bind_tip_df = function(tree = NULL, where, tip_label,
                       frac = 0.5, new_node_above = FALSE,
                       node_label = NULL, return_tree = TRUE,
                       tree_tbl = NULL, node_heights = NULL, use_castor = FALSE){
  if(frac > 1 | frac < 0) stop("frac must be between 0 and 1.")
  if(is.null(tree) & is.null(tree_tbl)) stop("No input tree, please assign one.")
  if(!is.null(tree_tbl) & is.null(node_heights)) stop("Please specify node_heights.")
  if(!is.null(tree)){
    # make sure node labels are assigned and unique
    if(is.null(tree$node.label) || any(duplicated(tree$node.label)))
        tree$node.label = paste0("N", 1:ape::Nnode(tree))
    tree = ape::makeLabel(tree, tips = FALSE, nodes = TRUE)
    tree_tbl = tidytree::as_tibble(tree)
    node_heights = ape::branching.times(tree)
  }
  if(is.null(node_label)) node_label = paste0("node", length(node_heights) + 1L)
  
  tree_tbl$isTip = !(tree_tbl$node %fin% tree_tbl$parent)
  if(!"is_tip" %in% names(tree_tbl)){
    tree_tbl$is_tip = tree_tbl$isTip
  } 
  tree_tbl_node = tree_tbl[tree_tbl$label == where, ] # original node
  if(!tree_tbl_node$isTip){
    if(use_castor){
      where_offspring = tree_tbl[tree_tbl$label %fin%
                                   castor::get_subtree_at_node(as_tree_isTip(tree_tbl), node = where)$subtree$tip.label, ]
      max_offspring = max(where_offspring$node)
    } else {
      where_offspring = tidytree::offspring(tree_tbl, .node = where)
      max_offspring = max(where_offspring$node[where_offspring$isTip])
    }
    node_height = node_heights[where]
  }
  node_orig = tree_tbl_node$node # original node number of target location
  at_root = tree_tbl_node$parent == tree_tbl_node$node
  tree_tbl_2 = tree_tbl # to hold results
  
  if(!tree_tbl_node$isTip) { # the target is a node
    if(at_root & new_node_above) message("New tip cannot be inserted above the root; attached instead.")
    if(new_node_above & !at_root){ # insert above the target location node
      # add node first, push the numbers of nodes if they are after the inserted one
      tree_tbl_2$parent[tree_tbl_2$parent >= node_orig] = tree_tbl_2$parent[tree_tbl_2$parent >= node_orig] + 1L
      tree_tbl_2$node[tree_tbl_2$node >= node_orig] = tree_tbl_2$node[tree_tbl_2$node >= node_orig] + 1L
      # insert the new node
      tree_tbl_2 = tibble::add_row(tree_tbl_2,
                                   parent = tree_tbl_2$parent[tree_tbl_2$label == where], 
                                   node = node_orig,
                                   branch.length = tree_tbl_node$branch.length * frac,
                                   label = node_label,
                                   is_tip = FALSE, isTip = FALSE)
      # add tip
      tree_tbl_2$parent = tree_tbl_2$parent + 1L # all nodes will be added 1
      tree_tbl_2$node = data.table::fifelse(tree_tbl_2$node > max_offspring, tree_tbl_2$node + 1L, tree_tbl_2$node)
      # update orginal node, which will have the inserted node as its parent
      tree_tbl_2$parent[tree_tbl_2$label == where] =
        tree_tbl_2$node[tree_tbl_2$label == node_label]
      tree_tbl_2$branch.length[tree_tbl_2$label == where] =
        tree_tbl_node$branch.length * (1 - frac)
      tree_tbl_2 = tibble::add_row(tree_tbl_2,
                                   parent = tree_tbl_2$node[tree_tbl_2$label == node_label],
                                   node = max_offspring + 1L,
                                   branch.length = tree_tbl_2$branch.length[tree_tbl_2$label == where] +
                                     node_height,
                                   label = tip_label, 
                                   is_tip = TRUE, isTip = TRUE)
    } else { # attach to a target node; no need to create a new one
      if(at_root){ # root
        tree_tbl_new = tibble::tibble(parent = node_orig + 1L, node = 1L,
                                      branch.length = node_height, label = tip_label,
                                      is_tip = TRUE, isTip = TRUE)
        tree_tbl$parent = tree_tbl$parent + 1L
        tree_tbl$node = tree_tbl$node + 1L
        tree_tbl_2 = dplyr::bind_rows(tree_tbl_new, tree_tbl)
      } else { # an internal node
        # tree_tbl_2 = tree_tbl
        tree_tbl_2$parent = tree_tbl_2$parent + 1L # all nodes will be added 1
        tree_tbl_2$node = data.table::fifelse(tree_tbl_2$node > max_offspring, tree_tbl_2$node + 1L, tree_tbl_2$node)
        tree_tbl_2 = tibble::add_row(tree_tbl_2,
                                     parent = tree_tbl_2$node[tree_tbl_2$label == where],
                                     node = max_offspring + 1L,
                                     branch.length = node_height,
                                     label = tip_label,
                                     is_tip = TRUE, isTip = TRUE)
      }
    }
  } else { # the target is a tip
    parent_orig = tree_tbl_node$parent
    # add node first, push the numbers of nodes if they are after the inserted one
    tree_tbl_2$parent[tree_tbl_2$parent > parent_orig] = tree_tbl_2$parent[tree_tbl_2$parent > parent_orig] + 1L
    tree_tbl_2$node[tree_tbl_2$node > parent_orig] = tree_tbl_2$node[tree_tbl_2$node > parent_orig] + 1L
    # insert the new node
    tree_tbl_2 = tibble::add_row(tree_tbl_2,
                                 parent = parent_orig, node = parent_orig + 1L,
                                 branch.length = tree_tbl_node$branch.length * frac,
                                 label = node_label,
                                 is_tip = FALSE, isTip = FALSE)
    # add tip
    tree_tbl_2$parent = tree_tbl_2$parent + 1L # all nodes will be added 1
    tree_tbl_2$node = data.table::fifelse(tree_tbl_2$node > node_orig, tree_tbl_2$node + 1L, tree_tbl_2$node)
    tree_tbl_2 = tibble::add_row(tree_tbl_2,
                                 parent = tree_tbl_2$node[tree_tbl_2$label == node_label],
                                 node = node_orig + 1L,
                                 branch.length = tree_tbl_node$branch.length * (1 - frac),
                                 label = tip_label,
                                 is_tip = TRUE, isTip = TRUE)
    # update orginal node, which will have the inserted node as its parent
    tree_tbl_2$parent[tree_tbl_2$label == where] =
      tree_tbl_2$node[tree_tbl_2$label == node_label]
    tree_tbl_2$branch.length[tree_tbl_2$label == where] =
      tree_tbl_2$branch.length[tree_tbl_2$label == tip_label]
    
    # # another way
    # tree_tbl_above = tree_tbl[1:node_orig, ]
    # tree_tbl_below = tree_tbl[-(1:node_orig), ]
    # # add tip
    # tree_tbl_above$parent = tree_tbl_above$parent + 1
    # tree_tbl_below$parent = tree_tbl_below$parent + 1
    # tree_tbl_below$node = tree_tbl_below$node + 1
    # tree_tbl_new = tibble::tibble(parent = tree_tbl_above$parent[tree_tbl_above$label == where],
    #                             node = node_orig + 1,
    #                             branch.length = tree_tbl_above$branch.length[tree_tbl_above$label == where],
    #                             label = tip_label)
    # tree_tbl_2 = dplyr::bind_rows(tree_tbl_above, tree_tbl_new, tree_tbl_below)
    # # add node
    # loc_in_df2 = which(tree_tbl_2$node == tree_tbl_new$parent)
    # tree_tbl_2$parent = data.table::fifelse(tree_tbl_2$parent > tree_tbl_new$parent, tree_tbl_2$parent + 1, tree_tbl_2$parent)
    # tree_tbl_2$parent[tree_tbl_2$label %fin% c(where, tip_label)] = tree_tbl_new$parent + 1
    # tree_tbl_2$node = data.table::fifelse(tree_tbl_2$node > tree_tbl_new$parent, tree_tbl_2$node + 1, tree_tbl_2$node)
    # tree_tbl_2 = dplyr::bind_rows(tree_tbl_2,
    #                             tibble::tibble(parent = tree_tbl_new$parent,
    #                                            node = tree_tbl_new$parent + 1,
    #                                            branch.length = tree_tbl_new$branch.length * frac,
    #                                            label = "")) %>%
    #   dplyr::arrange(node)
    # tree_tbl_2$branch.length[tree_tbl_2$label %fin% c(where, tip_label)] = tree_tbl_new$branch.length * (1 - frac)
  }
  
  # tree_tbl_2 = dplyr::arrange(tree_tbl_2, node)
  tree_tbl_2 = tree_tbl_2[order(tree_tbl_2$node), ]
  
  if(!inherits(tree_tbl_2, "tbl_tree")) class(tree_tbl_2) = c("tbl_tree", class(tree_tbl_2))
  
  if(return_tree) return(tidytree::as.phylo(tree_tbl_2))
  tree_tbl_2$isTip = NULL
  tree_tbl_2
}

context("Test different trees")

test_that("Test get tree for plants", {
  tt = get_tree(sp_list = test_plant_list, taxon = "plant", scenario = "at_basal_node")
  expect_s3_class(tt, "phylo")
  expect_equal(ape::Ntip(tt), nrow(test_plant_list))
  
  # tt = get_tree(test_plant_list, taxon = "plant", scenario = "S2")
  # expect_s3_class(tt, "phylo")
  # expect_equal(ape::Ntip(tt), nrow(test_plant_list))
  # 
  # tt = get_tree(test_plant_list, taxon = "plant", scenario = "S3")
  # expect_s3_class(tt, "phylo")
  # expect_equal(ape::Ntip(tt), nrow(test_plant_list))
  
  # test bind_tip with output file above
  tt2 = bind_tip(tree = tt, where = "Acer_cordatum", tip_label = "Test_sp")
  expect_equal(ape::Ntip(tt2), nrow(test_plant_list) + 1)
  tt3 = bind_tip(tt, where = "N37580", tip_label = "Test_sp")
  expect_equal(ape::Ntip(tt3), nrow(test_plant_list) + 1)
  tt4 = bind_tip(tt, where = "N37580", tip_label = "Test_sp", new_node_above = T)
  expect_equal(ape::Nnode(tt3), ape::Nnode(tt4) - 1)
  # message if add above root
  expect_message(bind_tip(tree = tt, where = "mrcaott2ott969", tip_label = "Test_sp", new_node_above = T))
})

test_that("Test get tree for fish", {
  expect_message(tt <- get_tree(sp_list = test_fish_list, taxon = "fish", scenario = "at_basal_node"))
  expect_s3_class(tt, "phylo")
  expect_equal(ape::Ntip(tt), nrow(test_plant_list) - 1)
})

test_that("Test get tree for bird", {
  tt = get_tree(sp_list = test_bird_list, taxon = "bird", show_grafted = T)
  expect_equal(ape::Ntip(tt[[1]]), nrow(test_bird_list))
})

test_that("Test get tree for mammal", {
  tt = get_tree(test_mammal_list, taxon = "mammal", show_grafted = T)
  expect_equal(ape::Ntip(tt[[1]]), nrow(test_mammal_list))
  # plot(ape::ladderize(tt))
})

test_that("Test user provided tree", {
  test_tree_sp2 = test_tree_sp[test_tree_sp != "Sorbus_sp"]
  expect_error(get_tree(sp_list = test_tree_sp, tree = test_tree, tree_by_user = T))
  tt = get_tree(sp_list = test_tree_sp, tree = test_tree, taxon = "plant", tree_by_user = T) # this should work
  tt = get_tree(sp_list = test_tree_sp2, tree = test_tree, tree_by_user = T) # this should work
  
  tt = get_tree(sp_list = test_tree_sp_df, tree = test_tree, taxon = "plant", show_grafted = T, tree_by_user = T)
})

# test_that("Test user provided tree and other taxon", {
#   load(rawConnection(RCurl::getBinaryURL("https://raw.githubusercontent.com/jinyizju/V.PhyloMaker/master/data/GBOTB.extended.rda")))
#   tt = get_tree(sp_list = sp_list_phylomatic, tree = GBOTB.extended, taxon = "other", show_grafted = T, tree_by_user = T)
# })

# Extracted from test_get_tree.R:24

# prequel ----------------------------------------------------------------------
context("Test different trees")

# test -------------------------------------------------------------------------
tt = get_tree(sp_list = test_plant_list, taxon = "plant", scenario = "at_basal_node")
expect_s3_class(tt, "phylo")
expect_equal(ape::Ntip(tt), nrow(test_plant_list))
tt2 = bind_tip(tree = tt, where = "Acer_cordatum", tip_label = "Test_sp")
expect_equal(ape::Ntip(tt2), nrow(test_plant_list) + 1)
tt3 = bind_tip(tt, where = "N37580", tip_label = "Test_sp")
expect_equal(ape::Ntip(tt3), nrow(test_plant_list) + 1)
tt4 = bind_tip(tt, where = "N37580", tip_label = "Test_sp", new_node_above = T)
expect_equal(ape::Nnode(tt3), ape::Nnode(tt4) - 1)
expect_message(bind_tip(tree = tt, where = "mrcaott2ott969", tip_label = "Test_sp", new_node_above = T))

# plants
test_plant_list = dplyr::bind_rows(
  # out of tree
  tibble::tibble(species = c("Meliosma_laui", "Acer_cordatum", "Fraxinus_mandshurica",
                             "Ormosia_pinnata", "Aglaia_dasyclada", "Claoxylon_indicum"),
                 genus = c("Meliosma", "Acer", "Fraxinus", "Ormosia", "Aglaia", "Claoxylon"),
                 family = c("Sabiaceae", "Sapindaceae", "Oleaceae", "Fabaceae", "Meliaceae", "Euphorbiaceae")),
  # in tree
  tibble::tibble(species = c("Carya_alba", "Carya_cordiformis", "Stewartia_sinensis", "Pterocarya_hupehensis",
                             "Ulmus_davidiana", "Diplospora_dubia"),
                 genus = c("Carya", "Carya", "Stewartia", "Pterocarya", "Ulmus", "Diplospora"),
                 family = c("Juglandaceae", "Juglandaceae", "Theaceae", "Juglandaceae", "Ulmaceae", "Rubiaceae"))
)

test_tree = ape::read.tree(text = "(((((((((Potentilla_intermedia:1.156097,Potentilla_gracilis:1.156097):9.169741,Potentilla_bipinnatifida:10.325839):10.466967,Potentilla_sterilis:20.792806):11.333216,Fragaria_virginiana:32.126022):6.026567,(((((Rosa_setigera:3.434279,Rosa_arkansana:3.434279):1.991106,Rosa_spinosissima:5.425386):0.993924,Rosa_acicularis:6.419309):6.730804,(Rosa_micrantha:0.188519,Rosa_canina:0.188519):12.961594):23.677485,(Agrimonia_gryposepala:4.730863,Agrimonia_parviflora:4.730863):32.096735):1.324992):6.47604,(Geum_urbanum:11.153303,Geum_rivale:11.153303):33.475326):1.449852,((((((((Rubus_semisetosus:3.10529,Rubus_glandicaulis:3.10529):0.957584,Rubus_steelei:4.062873):0.398394,Rubus_cuneifolius:4.461267):2.271628,Rubus_vagus:6.732895):0.023439,Rubus_superioris:6.756333):0.020104,(Rubus_multifer:4.287114,Rubus_elegantulus:4.287115):2.489323):0.025475,((Rubus_laciniatus:0.125186,Rubus_bifrons:0.125186):0.11903,Rubus_praecox:0.244216):6.557697):11.571849,(Rubus_parviflorus:14.85915,Rubus_odoratus:14.85915):3.514612):27.704718):16.915234,Filipendula_rubra:62.993714):13.354116,((((((((((Crataegus_dissona:0.081863,Crataegus_florifera:0.081863):0.49838,Crataegus_fulleriana:0.580243):1.434261,Crataegus_crus-galli:2.014503):0.177737,Crataegus_chrysocarpa:2.192239):0.051159,Crataegus_uniflora:2.243399):0.438351,Crataegus_persimilis:2.68175):4.697259,((Malus_toringo:6.472694,Malus_ioensis:6.472693):0.772012,Chaenomeles_speciosa:7.244705):0.134305):1.72489,Pyrus_communis:9.1039):0.313393,Amelanchier_spicata:9.417293):40.039557,((((((Prunus_pumila:6.189944,Prunus_nigra:6.189945):0.37487,Prunus_pensylvanica:6.564815):4.856241,(Prunus_padus:4.816999,Prunus_serotina:4.817):6.604057):29.162527,Sorbaria_sorbifolia:40.583584):2.924475,Rhodotypos_scandens:43.50806):2.718224,(Spiraea_X_bumalda:25.921582,Aruncus_dioicus:25.921583):20.304701):3.230567):26.89098);")

test_tree_sp = c("Rosa_sp", "Rubus_sp", "Amelanchier_sp", "Prunus_sp", "Sorbus_sp", "Potentilla_sp",
  "Potentilla_intermedia", "Potentilla_gracilis", "Potentilla_bipinnatifida", "Potentilla_sterilis", "Fragaria_virginiana", "Rosa_setigera", "Rosa_arkansana", "Rosa_spinosissima", "Rosa_acicularis", "Rosa_micrantha", "Rosa_canina", "Agrimonia_gryposepala", "Agrimonia_parviflora", "Geum_urbanum", "Geum_rivale", "Rubus_semisetosus", "Rubus_glandicaulis", "Rubus_steelei", "Rubus_cuneifolius", "Rubus_vagus", "Rubus_superioris", "Rubus_multifer", "Rubus_elegantulus", "Rubus_laciniatus", "Rubus_bifrons", "Rubus_praecox", "Rubus_parviflorus", "Rubus_odoratus", "Filipendula_rubra", "Crataegus_dissona", "Crataegus_florifera", "Crataegus_fulleriana", "Crataegus_crus-galli", "Crataegus_chrysocarpa", "Crataegus_uniflora", "Crataegus_persimilis", "Malus_toringo", "Malus_ioensis", "Chaenomeles_speciosa", "Pyrus_communis", "Amelanchier_spicata", "Prunus_pumila", "Prunus_nigra", "Prunus_pensylvanica", "Prunus_padus", "Prunus_serotina", "Sorbaria_sorbifolia", "Rhodotypos_scandens", "Spiraea_X_bumalda", "Aruncus_dioicus")

test_tree_sp_df = sp_list_df(test_tree_sp, "plant")
test_tree_sp_df$close_sp = NA
test_tree_sp_df$close_genus = NA
test_tree_sp_df$close_sp[1] = "Rosa_acicularis"
test_tree_sp_df$close_genus[4] = "Rosa"

setdiff(unique(sp_list_df(test_tree_sp)$genus), unique(sp_list_df(test_tree$tip.label)$genus))

sp_list_phylomatic = c("asteraceae/Achillea/Achillea_sp",
                       "berberidaceae/Achlys/Achlys_sp",
                       "ranunculaceae/Aconitum/Aconitum_sp",
                       "ranunculaceae/Actaea/Actaea_rubra",
                       "asteraceae/Adenocaulon/Adenocaulon_sp")

# fish
test_fish_list = tibble::tibble(
  species = c("Serrasalmus_geryi", "Careproctus_reinhardti", "Gobiomorphus_coxii", 
              "Periophthalmus_barbarus", "Prognichthys_glaphyrae", "Barathronus_bicolor", 
              "Knipowitschia_croatica", "Rhamphochromis_lucius", "Neolissochilus_tweediei", 
              "Haplochromis_nyanzae", "Astronesthes_micropogon", "Sanopus_reticulatus"),
  genus = c("Serrasalmus", "Careproctus", "Gobiomorphus", "Periophthalmus",
            "Prognichthys", "Barathronus", "Knipowitschia", "Rhamphochromis", 
            "Neolissochilus", "Haplochromis", "Astronesthes", "Sanopus"),
  family = c("Serrasalmidae", "Liparidae", "Eleotridae", "Gobiidae", 
             "Exocoetidae", "Aphyonidae", "Gobiidae", "Cichlidae", 
             "Cyprinidae", "Cichlidae", "Stomiidae", "Batrachoididae")
)

# bird
test_bird_list = c("Brachypteryx_major", "Asthenes_perijana", "Ciridops_anna", 
                   "Leiothlypis_ruficapilla", "Reinwardtoena_reinwardti",
                   "Garrulax_caerulatus", "Buteo_rufofuscus", "Sylvia_mystacea",
                   "Telophorus_viridis", "Trachyphonus_darnaudii") %>% 
  sp_list_df(taxon = "bird")

test_mammal_list = c("Necromys_temchuki", "Hipposideros_beatus", "Ateles_marginatus",
                     "Protemnodon_anak", "Mus_musculus", "Echymipera_sp", "Echymipera_davidi") %>% 
  sp_list_df(taxon = "mammal")

#' Derive a phylogeny from a mega-tree
#' 
#' For a list of species, generate a phylogeny from a provided mega-tree. If a species is
#' not in the mega-tree, it will be grafted to the mega-tree with three scenarioes.
#' 
#' @inheritParams get_tree
#' @return A phylogeny for the species required, with class `phylo`.
#' @importFrom stats setNames
#' @export
#' 
get_one_tree = function(sp_list, tree, taxon, 
                        scenario = c("at_basal_node", "random_below_basal"), 
                        show_grafted = FALSE,
                        tree_by_user = FALSE,
                        .progress = "text", dt = TRUE) {
  if(tree_by_user & all(!grepl("_", tree$tip.label)))
    stop("Please change the tree's tip labels to be the format of genus_sp.")
  if(tree_by_user) tree = rm_stars(tree)
  tree_genus = unique(gsub("^([-A-Za-z]*)_.*$", "\\1", tree$tip.label))
  
  sp_list = sp_list_df(unique(sp_list)) # remove duplication and prep genus, family
  
  all_genus_in_tree = all(unique(sp_list$genus) %fin% tree_genus)
  # if TRUE, no taxon is required
  if(!all_genus_in_tree){
    if((!"family" %fin% names(sp_list))){ # no family info
      if(missing(taxon) | is.null(taxon)) 
        stop("Please specify `taxon` as not all genus are in the tree.")
      sp_list = sp_list_df(sp_list$species, taxon) # add family information
    }
  }
  
  # add new classification data to classification data frame
  if(!is.null(taxon)){
    if(!taxon %fin% rtrees::taxa_supported & !all_genus_in_tree){
      new_cls = unique(dplyr::select(sp_list, genus, family))
      new_cls$taxon = taxon
      classifications <- dplyr::bind_rows(rtrees::classifications, new_cls)
    }
  } 
  
  sp_out_tree = sp_list[!sp_list$species %fin% tree$tip.label, ]
  
  # some tree tips has genus_sp_subsp
  subsp_in_tree = grep("^.*_.*_.*$", x = tree$tip.label, value = TRUE)
  if(length(subsp_in_tree)){
    sp_out_tree = dplyr::mutate(sp_out_tree, re_matched = NA, matched_name = NA)
    for(i in 1:length(sp_out_tree$species)){
      name_in_tree = grep(paste0("^", sp_out_tree$species[i], "_"), x = subsp_in_tree,
                          ignore.case = TRUE, value = TRUE)
      # avoid Allium_sp. match things like Allium_splendens
      # cat(i, name_in_tree, "\n")
      if(length(name_in_tree)) {
        sp_out_tree$re_matched[i] = TRUE
        sp_out_tree$matched_name[i] = sample(name_in_tree, 1)
        # rename the tree tip 
        tree$tip.label[tree$tip.label == sp_out_tree$matched_name[i]] = sp_out_tree$species[i]
        if(!is.null(tree$genus_family_root)) {
          tree$genus_family_root$only_sp[tree$genus_family_root$only_sp == 
                                           sp_out_tree$matched_name[i]] = 
            sp_out_tree$species[i]
        }
      }
    }
    sp_out_tree = dplyr::distinct(sp_list[!sp_list$species %fin% tree$tip.label, ])
  }
  
  close_sp_specified = close_genus_specified = FALSE
  if("close_sp" %fin% names(sp_out_tree)) {
    close_sp_specified = TRUE
    # in case of white spaces in names
    sp_out_tree$close_sp = cap_first_letter(gsub(" +", "_", sp_out_tree$close_sp))
  }
  if("close_genus" %fin% names(sp_out_tree)) close_genus_specified = TRUE
  
  if(nrow(sp_out_tree) == 0){
    message("Wow, all species are already in the mega-tree!")
    tree_sub = ape::drop.tip(tree, setdiff(tree$tip.label, sp_list$species))
    return(tree_sub)
  }
  
  if(tree_by_user){
    if(!is.null(tree$genus_family_root)) 
      warning("The phylogeny has basal node information, are you sure this is an user provided tree?")
    if(all_genus_in_tree){
      tree = add_root_info(tree, process_all_tips = FALSE,
                           genus_list = unique(sp_out_tree$genus), show_warning = FALSE)
    } else { # some genus not in the tree
      if(missing(taxon)) stop("Please specify `taxon`.")
      message("Not all genus can be found in the phylogeny.")
      if(is.null(tree$genus_family_root)){
        warning("For user provided phylogeny, without a classification for all genus of species in the phylogeny,
              it is unlikely to find the most recent ancestor for genus and family; here we proceed the phylogeny
              by adding root information for genus and family that can be found in the phylogeny or species list but
              we recommend to prepare the phylogeny using `add_root_info()` with a classification
              data frame with all tips first.", call. = FALSE, immediate. = TRUE)
      }
      genus_not_in_tree = dplyr::filter(sp_out_tree, !genus %fin% tree_genus)
      
      # add root information for species not in the tree
      tree = add_root_info(
        tree, 
        classification = if(is.null(taxon) & 
                            inherits(sp_list, "data.frame") & 
                            all(c("genus", "family") %fin% names(sp_list))){
          unique(sp_list[, c("genus", "family")])
        } else {
          if((!taxon %fin% rtrees::taxa_supported) & !all_genus_in_tree){
            unique(classifications[classifications$taxon == taxon, ])
          } else {
            unique(rtrees::classifications[rtrees::classifications$taxon == taxon, ])
          }
        }, 
        process_all_tips = FALSE,
        genus_list = if(length(setdiff(sp_out_tree$genus, genus_not_in_tree$genus))) setdiff(sp_out_tree$genus, genus_not_in_tree$genus) else NULL,
        family_list = if(nrow(genus_not_in_tree) > 0) unique(genus_not_in_tree$family) else NULL,
        show_warning = FALSE)
    }
  }
  
  scenario = match.arg(scenario)

  if(is.null(tree$genus_family_root))
    stop("Did you use your own phylogeny? If so, please set `tree_by_user = TRUE`.")
  sp_out_tree$status = ""

  # Use plain integer/numeric vectors instead of a tibble throughout the loop.
  # phylo_to_vecs() is O(Ntip) and avoids all tidytree/[[.tbl_df overhead.
  vecs <- phylo_to_vecs(tree)
  # O(1) label → row-index lookup updated incrementally after each graft.
  label_env <- list2env(
    setNames(as.list(seq_along(vecs$label)), vecs$label),
    hash = TRUE, parent = emptyenv())
  # O(1) node-height (branching time) lookup; new nodes added via [[ <- ]] .
  node_hts_env <- list2env(as.list(ape::branching.times(tree)), hash = TRUE, parent = emptyenv())
  n_internal_initial <- ape::Nnode(tree)
  n_nodes_added <- 0L

  all_eligible_nodes = unique(c(tree$genus_family_root$basal_node,
                                tree$genus_family_root$root_node))

  n_spp_to_show_progress = 200
  if(nrow(sp_out_tree) > n_spp_to_show_progress){
    progress <- create_progress_bar(.progress)
    progress$init(nrow(sp_out_tree))
    on.exit(progress$term())
  }

  # For at_basal_node: pre-allocate graft-parameter arrays; call graft_all_cpp()
  # once after the loop (one C++ allocation instead of N tibble round-trips).
  n_sp <- nrow(sp_out_tree)
  if(scenario == "at_basal_node") {
    gb_where <- character(n_sp)
    gb_nlbl  <- character(n_sp)   # "" = no new internal node needed
    gb_frac  <- rep(0.5, n_sp)
    gb_above <- logical(n_sp)
    gb_ht    <- numeric(n_sp)
    gb_valid <- logical(n_sp)
  }

  for(i in 1:n_sp){
    if(nrow(sp_out_tree) > n_spp_to_show_progress)
      progress$step()

    where_loc_i = where_loc_i2 = NA

    if(close_sp_specified){
      if(!is.na(sp_out_tree$close_sp[i]) &
         sp_out_tree$close_sp[i] %fin% tree$tip.label){
        where_loc_i = sp_out_tree$close_sp[i]
      }
    }

    if(close_genus_specified){
      if(!is.na(sp_out_tree$close_genus[i]) &
         sp_out_tree$close_genus[i] != "" &
         sp_out_tree$close_genus[i] %fin% tree_genus){
        sp_out_tree$genus[i] = sp_out_tree$close_genus[i]
        where_loc_i2 = sp_out_tree$close_genus[i]
      } else {
        if(!is.na(sp_out_tree$close_genus[i]))
          warning("The genus specified for ", sp_out_tree$species[i],
                  " is not in the phylogeny.")
      }
    }

    if(!all_genus_in_tree & is.na(where_loc_i) & is.na(where_loc_i2)){
      if(is.na(sp_out_tree$family[i]) |
         !sp_out_tree$family[i] %fin% tree$genus_family_root$family){
        sp_out_tree$status[i] = "No co-family species in the mega-tree"
        next()
      }
    }

    node_label_new = NULL
    add_above_node = FALSE
    fraction = 1/2
    new_genus_entry_added = FALSE  # set TRUE when we register a new genus in family fallback

    if(sp_out_tree$genus[i] %fin% tree$genus_family_root$genus |
       !is.na(where_loc_i2) | !is.na(where_loc_i)){
      sp_out_tree$status[i] = "*"
      # tree has species in the same genus
      idx_row = which(tree$genus_family_root$genus == sp_out_tree$genus[i])
      root_sub = tree$genus_family_root[idx_row, ]
      if(isTRUE(root_sub$n_spp == 1) | !is.na(where_loc_i)) { # but only 1 species in this genus
        if(!is.na(where_loc_i)){# a close sp specified
          where_loc = where_loc_i
          wi <- label_env[[where_loc_i]]
          new_ht = vecs$branch.length[wi] * (1 - fraction)
          n_nodes_added <- n_nodes_added + 1L
          node_label_new = paste0("N", n_internal_initial + n_nodes_added)
          node_hts_env[[node_label_new]] <- new_ht
          all_eligible_nodes = c(all_eligible_nodes, node_label_new)
          add_above_node = TRUE
          if(!sp_out_tree$genus[i] %fin% tree$genus_family_root$genus){
            # this is a new genus that has not in the tree
            par_num <- vecs$parent[wi]
            # For the initial tree vecs$node[j]==j, so par_num IS the row index.
            # Use match() to be safe across random_below_basal grafts.
            par_row <- match(par_num, vecs$node)
            par_label <- vecs$label[par_row]
            nh_par <- node_hts_env[[par_label]]
            if(is.null(nh_par)) nh_par <- 0.0
            tree$genus_family_root = tibble::add_row(
              tree$genus_family_root,
              family = sp_out_tree$family[i],
              genus = sp_out_tree$genus[i],
              basal_node = node_label_new,
              basal_time = new_ht,
              root_node = par_label,
              root_time = nh_par,
              n_genus = 1, n_spp = 1, only_sp = sp_out_tree$species[i])
            idx_row = nrow(tree$genus_family_root)
          }
        } else {
          where_loc = root_sub$only_sp
          # the new tip will be bind to this species, in the half of its branch length (default frac)
          new_ht = root_sub$basal_time * (1 - fraction)
          n_nodes_added <- n_nodes_added + 1L
          node_label_new = paste0("N", n_internal_initial + n_nodes_added)
          node_hts_env[[node_label_new]] <- new_ht
          all_eligible_nodes = c(all_eligible_nodes, node_label_new)
          add_above_node = TRUE
          tree$genus_family_root$only_sp[idx_row] = NA # now will be more than 1 sp in this genus
          tree$genus_family_root$basal_node[idx_row] = node_label_new
          tree$genus_family_root$basal_time[idx_row] = unname(new_ht)
        }
      } else { # more than 1 species in the genus
        where_loc = root_sub$basal_node # scenarioes 1 and 3, no new node added
        if(scenario == "random_below_basal"){ # randomly select a node in the genus and attach to it, no new node added
          wi <- label_env[[where_loc]]
          where_loc_node = vecs$node[wi]
          sub_idx = fast_internal_offspring_cpp(vecs$parent, vecs$node, vecs$is_tip, where_loc_node)
          if(length(sub_idx) > 0){
            sub_labels <- vecs$label[sub_idx]
            sub_bls    <- vecs$branch.length[sub_idx]
            potential_locs = c(where_loc, sub_labels)
            bls = sub_bls
            names(bls) = sub_labels
            bls = c(root_sub$root_time - root_sub$basal_time, bls)
            names(bls)[1] = root_sub$basal_node
            prob = bls/sum(bls)
            where_loc = sample(potential_locs, 1, prob = prob)
          }
        }
      }
    } else {
      # no species in the same genus in the tree, go up to family node
      sp_out_tree$status[i] = "**"
      idx_row = which(tree$genus_family_root$family == sp_out_tree$family[i] &
                        is.na(tree$genus_family_root$genus))
      root_sub = tree$genus_family_root[idx_row, ]

      if(root_sub$n_spp == 1) { # but only 1 species in this family
        where_loc = root_sub$only_sp
        # the new tip will be bind to this species, in the half of its branch length (default frac)
        new_ht = root_sub$basal_time * (1 - fraction)
        n_nodes_added <- n_nodes_added + 1L
        node_label_new = paste0("N", n_internal_initial + n_nodes_added)
        node_hts_env[[node_label_new]] <- new_ht
        all_eligible_nodes = c(all_eligible_nodes, node_label_new)
        add_above_node = TRUE
        tree$genus_family_root = tibble::add_row(tree$genus_family_root,
                                                 family = sp_out_tree$family[i],
                                                 genus = sp_out_tree$genus[i],
                                                 basal_node = node_label_new,
                                                 basal_time = unname(new_ht),
                                                 root_node = root_sub$basal_node,
                                                 root_time = root_sub$basal_time,
                                                 n_genus = 1,
                                                 n_spp = 1,
                                                 only_sp = sp_out_tree$species[i]
        )
      } else { # more than 1 species; can be the same genus or different genus
        where_loc = root_sub$basal_node # for scenario 1, no new node added
        if(scenario == "random_below_basal"){ # randomly select a node in the family, no new node added
          wi <- label_env[[where_loc]]
          where_loc_node = vecs$node[wi]
          sub_idx = fast_internal_offspring_cpp(vecs$parent, vecs$node, vecs$is_tip, where_loc_node)
          if(length(sub_idx) > 0){
            sub_labels <- vecs$label[sub_idx]
            sub_bls    <- vecs$branch.length[sub_idx]
            # only bind to genus/family basal node, not within genus nodes
            potential_locs = intersect(c(where_loc, sub_labels), all_eligible_nodes)
            pl_in_sub <- potential_locs[potential_locs %fin% sub_labels]
            bls = sub_bls[match(pl_in_sub, sub_labels)]
            names(bls) = pl_in_sub
            bls = c(root_sub$root_time - root_sub$basal_time, bls)
            names(bls)[1] = root_sub$basal_node
            prob = bls/sum(bls)
            where_loc = sample(potential_locs, 1, prob = prob)
          }
        }
        # if(scenario == "at_or_above_basal"){ # insert new node and bind tip above family basal node
        #   add_above_node = TRUE
        #   if(2 * root_sub$root_time / 3 > root_sub$basal_time){
        #     fraction = (2 * root_sub$root_time / 3 - root_sub$basal_time) /
        #       (root_sub$root_time - root_sub$basal_time)
        #   }
        #   new_ht = unname(root_sub$basal_time + (root_sub$root_time - root_sub$basal_time) * (1 - fraction))
        #   # here is the node height, but in bind_tip, it is the length between a parent and a node,
        #   # thus 1 - fraction
        #   n_nodes_added <- n_nodes_added + 1L
        #   node_label_new = paste0("N", n_internal_initial + n_nodes_added)
        #   node_hts_env[[node_label_new]] <- new_ht
        #   all_eligible_nodes = c(all_eligible_nodes, node_label_new)
        #   tree$genus_family_root$basal_node[idx_row] = node_label_new # new basal node
        #   tree$genus_family_root$basal_time[idx_row] = new_ht
        #   tree$genus_family_root = tibble::add_row(tree$genus_family_root,
        #                                            family = sp_out_tree$family[i],
        #                                            genus = sp_out_tree$genus[i],
        #                                            basal_node = node_label_new,
        #                                            basal_time = new_ht,
        #                                            root_node = root_sub$root_node,
        #                                            root_time = root_sub$root_time,
        #                                            n_genus = 1,
        #                                            n_spp = 1,
        #                                            only_sp = sp_out_tree$species[i])
        # }
        # Register a new genus entry only when there are additional missing
        # species from the same new genus, so they form a polytomy clade above
        # the first attached species rather than scattering at the family basal node.
        if (i < n_sp) {
          genus_i <- sp_out_tree$genus[i]
          remaining_genera <- sp_out_tree$genus[(i + 1L):n_sp]
          n_same_genus_remaining <- sum(
            remaining_genera == genus_i &
            !remaining_genera %fin% tree$genus_family_root$genus
          )
          if (n_same_genus_remaining > 0L) {
            tree$genus_family_root = tibble::add_row(
              tree$genus_family_root,
              family     = sp_out_tree$family[i],
              genus      = genus_i,
              basal_node = root_sub$basal_node,
              basal_time = root_sub$basal_time,
              root_node  = root_sub$root_node,
              root_time  = root_sub$root_time,
              n_genus    = 1L,
              n_spp      = 1L,
              only_sp    = sp_out_tree$species[i]
            )
            new_genus_entry_added = TRUE
          }
        }
      }
      # update genus number
      tree$genus_family_root$n_genus[idx_row] = tree$genus_family_root$n_genus[idx_row] + 1
    }

    # ---- Graft ----
    if(scenario == "at_basal_node") {
      # Collect parameters; the single graft_all_cpp() call happens after the loop.
      nh_val <- node_hts_env[[where_loc]]
      gb_where[i] <- where_loc
      gb_nlbl[i]  <- if (!is.null(node_label_new)) node_label_new else ""
      gb_frac[i]  <- fraction
      gb_above[i] <- add_above_node
      gb_ht[i]    <- if (!is.null(nh_val)) nh_val else 0.0
      gb_valid[i] <- TRUE
    } else {
      # random_below_basal: update vecs in place after each graft.
      # bind_tip_core_cpp() operates on plain vectors — no tibble allocation per step.
      nh_val <- node_hts_env[[where_loc]]
      if(is.null(nh_val)) nh_val <- 0.0
      res <- bind_tip_core_cpp(
        parent        = vecs$parent,
        node          = vecs$node,
        branch_length = vecs$branch.length,
        label         = vecs$label,
        is_tip        = vecs$is_tip,
        where         = where_loc,
        tip_label     = sp_out_tree$species[i],
        node_label_str = if (!is.null(node_label_new)) node_label_new else "",
        frac          = fraction,
        new_node_above = add_above_node,
        node_height   = nh_val
      )
      n_old <- length(vecs$label)
      vecs$parent        <- res$parent
      vecs$node          <- res$node
      vecs$branch.length <- res$`branch.length`
      vecs$label         <- res$label
      vecs$is_tip        <- res$is_tip
      for (j in (n_old + 1L):length(vecs$label)) label_env[[vecs$label[j]]] <- j
      if (new_genus_entry_added) {
        # Correct basal_time to the actual branch length of the first species after
        # random placement (may differ from the family basal time used as initial estimate).
        tree$genus_family_root$basal_time[nrow(tree$genus_family_root)] <-
          vecs$branch.length[label_env[[sp_out_tree$species[i]]]]
      }
    }

    # update n_spp in tree$genus_family_root
    tree$genus_family_root$n_spp[idx_row] = tree$genus_family_root$n_spp[idx_row] + 1
  }

  # ---- Post-loop: materialise the final tree_df ----
  if(scenario == "at_basal_node") {
    valid_i <- which(gb_valid)
    if(length(valid_i) > 0) {
      result <- graft_all_cpp(
        parent0        = vecs$parent,
        node0          = vecs$node,
        bl0            = vecs$branch.length,
        label0         = vecs$label,
        is_tip0        = vecs$is_tip,
        where_labels   = gb_where[valid_i],
        new_tip_labels = sp_out_tree$species[valid_i],
        new_node_labels = gb_nlbl[valid_i],
        fracs          = gb_frac[valid_i],
        above_flags    = gb_above[valid_i],
        node_heights_vec = gb_ht[valid_i]
      )
      tree_df <- tibble::as_tibble(result)
    } else {
      tree_df <- tibble::as_tibble(vecs)
    }
  } else {
    tree_df <- tibble::as_tibble(vecs)
  }
  if(!inherits(tree_df, "tbl_tree")) class(tree_df) <- c("tbl_tree", class(tree_df))

  tree_df = dplyr::arrange(tree_df, node)
  
  # if(nrow(sp_out_tree) > 100){
  #   close(progbar)
  # }
  
  if(any(sp_out_tree$status == "*")) {
    message("\n", sum(sp_out_tree$status == "*"), " species added at genus level (*) \n")
  }
  
  if(any(sp_out_tree$status == "**")) {
    message(sum(sp_out_tree$status == "**"), " species added at family level (**) \n")
  }
  
  tree_sub = castor::get_subtree_with_tips(tidytree::as.phylo(tree_df), sp_list$species)$subtree
  
  # in case of non ultrametric
  if(ape::is.ultrametric(tree) & !ape::is.ultrametric(tree_sub)){
    ntips = ape::Ntip(tree_sub)
    ages_tips = castor::get_all_distances_to_root(tree_sub)[1 : ntips]
    ages_diff = round(ages_tips - stats::median(ages_tips), 3) # the most common age should be the median
    tree_sub_df = tidytree::as_tibble(tree_sub)
    stopifnot(all(!tree_sub_df$node[1:ntips] %in% tree_sub_df$parent))
    tree_sub_df$branch.length[1:ntips] = pmax(0, tree_sub_df$branch.length[1:ntips] - ages_diff)
    tree_sub = tidytree::as.phylo(tree_sub_df)
    # ultrametric
    tree_sub = castor::extend_tree_to_height(tree_sub)$tree
  }
  
  # add trailing *
  grafted = sp_out_tree[sp_out_tree$status %fin% c("*", "**"), ]
  grafted$sp2 = paste0(grafted$species, grafted$status)
  wid = which(tree_sub$tip.label %fin% grafted$species)
  tree_sub$tip.label[wid] = dplyr::left_join(tibble::tibble(species = tree_sub$tip.label[wid]),
                                             grafted, by = "species")$sp2
  
  graft_status = tibble::tibble(tip_label = tree_sub$tip.label)
  graft_status$species = gsub("\\*", "", graft_status$tip_label)
  graft_status$status = ifelse(grepl("\\*{2}$", graft_status$tip_label), "grafted at family level",
                      ifelse(grepl("[^*]\\*{1}$", graft_status$tip_label), "grafted at genus level",
                             "exisiting species in the megatree"))
  
  if(any(sp_out_tree$status == "No co-family species in the mega-tree")){
    sp_no_family = sp_out_tree$species[sp_out_tree$status == "No co-family species in the mega-tree"]
    message(length(sp_no_family), " species have no co-family species in the mega-tree, skipped\n(if you know their family, prepare and edit species list with `rtrees::sp_list_df()` may help): \n",
            paste(sp_no_family, collapse = ", "))
    graft_status = dplyr::bind_rows(graft_status, 
                                    data.frame(species = sp_no_family, 
                                                  status = rep("skipped as no co-family in the megatree", 
                                                               length(sp_no_family)))
    )
  }
  
  if(!show_grafted){
    tree_sub = rm_stars(tree_sub)
    graft_status$tip_label = gsub("\\*", "", graft_status$tip_label)
  }
  
  tree_sub$graft_status = graft_status
  
  tree_sub = ape::ladderize(tree_sub)
  
  return(tree_sub)
}

#' Faster match of vectors
#'
#' See \code{fastmatch::\link[fastmatch:fmatch]{\%fin\%}} for details.
#'
#' @name %fin%
#' @rdname fmatch
#' @keywords internal
#' @importFrom fastmatch %fin%
#' @import data.table
NULL

#' #' Faster match of character vectors
#' #'
#' #' See \code{data.table::\link[data.table:chmatch]{\%chin\%}} for details.
#' #'
#' #' @name %chin%
#' #' @rdname chmatch
#' #' @keywords internal
#' #' @export
#' #' @importFrom data.table %chin%
#' NULL


#' Captize the first letter of a word.
#' @param x A word.
#' @return The same word with the first letter captized.
#' @noRd
cap_first_letter <- function(x) {
  sub("^([a-z])", "\\U\\1", x, perl = TRUE)
}

# BFS to find the max tip node-number among all descendants of root_node.
# Avoids converting the whole table to a phylo object on every bind_tip call.
fast_max_tip <- function(parent_vec, node_vec, is_tip_vec, root_node) {
  non_root <- which(parent_vec != node_vec)
  children <- split(non_root, parent_vec[non_root])
  frontier <- as.character(root_node)
  max_tip  <- 0L
  while (length(frontier) > 0L) {
    ch_idx <- unlist(children[frontier], use.names = FALSE)
    if (!length(ch_idx)) break
    is_tip    <- is_tip_vec[ch_idx]
    tip_nodes <- node_vec[ch_idx][is_tip]
    if (length(tip_nodes)) max_tip <- max(max_tip, max(tip_nodes))
    frontier <- as.character(node_vec[ch_idx][!is_tip])
  }
  max_tip
}

# BFS returning row indices of all internal-node descendants of root_node.
# Replaces tidytree::offspring() in random_below_basal sampling.
fast_internal_offspring <- function(parent_vec, node_vec, is_tip_vec, root_node) {
  non_root <- which(parent_vec != node_vec)
  children <- split(non_root, parent_vec[non_root])
  frontier   <- as.character(root_node)
  result_idx <- integer(0)
  while (length(frontier) > 0L) {
    ch_idx <- unlist(children[frontier], use.names = FALSE)
    if (!length(ch_idx)) break
    is_tip     <- is_tip_vec[ch_idx]
    result_idx <- c(result_idx, ch_idx[!is_tip])
    frontier   <- as.character(node_vec[ch_idx][!is_tip])
  }
  result_idx
}

if (getRversion() >= "2.15.1") {
  utils::globalVariables(c(
    ".", "isTip", "is_tip", "node", "tree_bee", "tree_butterfly",
    "tree_fish", "tree_plant_otl", "classifications",
    "tree_bird_ericson", "tree_mammal", "taxon",
    "family", "genus", "species", "grp", "parent",
    "root_node", "basal_node", "taxa_supported"
  ))
}

#' Taxonomic groups supported
#'
#' @description Supported taxonomic groups with mega-trees provided in the \pkg{megatrees} package.
#'
"taxa_supported"

#' Convert a vector of species names to a data frame
#'
#' @param sp_list A string vector or a data frame with at least one column named "species".
#' @param taxon The taxon group of this species list. If not specified, only species and
#' genus will be returned.
#' @return A data frame with columns: species, genus, and family (if `taxon` is specified).
#' @export
#' @examples
#' sp_list_df(
#'   sp_list = c("Serrasalmus_geryi", "Careproctus_reinhardti", "Gobiomorphus_coxii"),
#'   taxon = "fish"
#' )
sp_list_df <- function(sp_list, taxon) {
  if (!is.vector(sp_list, mode = "character") &
    !inherits(sp_list, "data.frame")) {
    stop("`sp_list` must either be a string vector or a data frame")
  }

  if (is.vector(sp_list, mode = "character")) { # vector
    # phylomatic format
    if (all(grepl(pattern = "[/]", x = sp_list))) { # phylomatic format
      sp_list_sep <- strsplit(sp_list, split = "/")
      sp_list <- tibble::tibble(
        species = cap_first_letter(sapply(
          sp_list_sep,
          function(x) gsub(" +", "_", x[3])
        )),
        genus = cap_first_letter(sapply(sp_list_sep, function(x) x[2])),
        family = cap_first_letter(sapply(sp_list_sep, function(x) x[1]))
      )
      return(sp_list)
    }

    sp_list <- unique(cap_first_letter(gsub(" +", "_", sp_list)))
    out <- tibble::tibble(
      species = sp_list,
      genus = gsub("^([-A-Za-z]*)_.*$", "\\1", sp_list)
    )
    if (missing(taxon)) {
      return(out)
    }
  } else { # data frame
    if (!"species" %in% names(sp_list)) {
      stop("`sp_list` must has at least one column named species.")
    }
    sp_list$species <- cap_first_letter(gsub(" +", "_", sp_list$species)) # just in case
    if ("genus" %in% names(sp_list)) {
      sp_list$genus <- cap_first_letter(sp_list$genus)
      if (missing(taxon)) {
        return(sp_list)
      }
    } else { # no genus column
      sp_list$genus <- gsub("^([-A-Za-z]*)_.*$", "\\1", sp_list$species)
      if ("family" %in% names(sp_list) | missing(taxon)) { # already have family, nothing to do
        return(sp_list)
      }
    }
    out <- sp_list
  }

  if (!taxon %fin% rtrees::taxa_supported) {
    stop(
      "Sorry but only the following taxon groups are supported: ",
      paste(rtrees::taxa_supported, collapse = ", "),
      "\n You need to prepare the species list data frame by yourself,",
      "\n which should have at least three columns: species, genus, family"
    )
  }
  # utils::data("classifications", envir = environment())
  clsf <- rtrees::classifications[rtrees::classifications$taxon == taxon, ]
  if (any(!out$genus %fin% clsf$genus)) {
    warning("The following genus are not in our classification database: ",
      paste(setdiff(out$genus, clsf$genus), collapse = ", "),
      call. = FALSE
    )
    # not necessary to be a problem, if tree has same genus species,
    # then no family info needed
  }
  if (mean(out$genus %fin% clsf$genus) < 0.8) {
    warning("Are you sure that you specified the right taxon group?", call. = FALSE)
  }
  out <- dplyr::left_join(out, clsf, by = "genus")
  out$taxon <- NULL
  unique(out)
}

#' Add genus and family basal/root node information to a phylogeny
#'
#' Based on the classification of tips, find where is the basal and root node for
#' each genus and each family. Such information can be later used to graft new
#' tips onto the phylogeny. This function can be used to process a user provided
#' tree.
#'
#' @param tree A phylogeny with class "phylo".
#' @param classification A data frame of 2 columns: genus, family. It should include
#' all genus the tips of the tree belong to.
#' @param process_all_tips Whether to find basal nodes for all tips? Default is `TRUE`.
#' @param genus_list An optinoal subset list of genus to find root information.
#' @param family_list An optinoal subset list of family to find root information.
#' This should be for species that do not have co-genus in the tree.
#' @param show_warning Whether to print warning information about non-monophyletic clades or not.
#' @return A phylogeny with basal nodes information attached.
#' @export
#'
add_root_info <- function(tree, classification, process_all_tips = TRUE,
                          genus_list = NULL, family_list = NULL,
                          show_warning = FALSE) {
  if (is.null(tree$node.label)) {
    tree$node.label <- paste0("N", 1:ape::Nnode(tree))
  }
  tree <- ape::makeLabel(tree, tips = FALSE, node = TRUE)
  if (any(ww <- grepl("^[0-9]*$", tree$node.label))) {
    tree$node.label[ww] <- paste0("N", tree$node.label[ww])
  }
  if (any(duplicated(tree$node.label))) {
    tree$node.label <- make.unique(tree$node.label, sep = "_")
  }
  tips <- tibble::tibble(
    species = tree$tip.label,
    genus = gsub("^([-A-Za-z]*)_.*$", "\\1", tree$tip.label)
  )

  if (process_all_tips) {
    if (!is.null(genus_list)) {
      stop("When `process_all_tips = TRUE`, `genus_list` must be NULL.")
    }
    if (any(!tips$genus %fin% classification$genus) & show_warning) {
      warning("Some genus are not in the classification.")
    }
    # add family information
    tips <- dplyr::left_join(tips, classification, by = "genus")
  } else { # only need for a subset of genus/family
    if (!is.null(family_list)) { # both genus and family
      # add family information
      family_list <- family_list[!is.na(family_list)]
      # classification will be filtered to the taxon in `get_one_tree()`
      # need family infor for genus in the phylogeny !!
      tips <- dplyr::left_join(tips, classification, by = "genus")
      if (all(is.na(tips$family))) {
        stop("No tips in the user provided tree can get family information from classification,
             please use `add_root_info()` to prepare the tree first.")
      }
      if (any(!family_list %fin% tips$family) & show_warning) {
        warning("Some family_list are not in the tree; these species will be ignored.")
      }
      tips_family <- tips[tips$family %fin% family_list, ]
      # if(any(!genus_list %fin% tips$genus) & show_warning)
      #   warning("Some genus_list are not in the tree.")
      ## these genus' family will be in the family_list
      tips_genus <- tips[tips$genus %fin% genus_list, ]
      tips <- unique(dplyr::bind_rows(tips_genus, tips_family))
    } else { # only genus
      if (is.null(genus_list)) {
        stop("When `process_all_tips = FALSE`, `genus_list` must be specified.")
      }
      if (any(!genus_list %fin% tips$genus) & show_warning) {
        warning("Some genus_list are not in the tree.")
      }
      tips <- tips[tips$genus %fin% genus_list, ] # no family column
    }
  }

  if ("family" %fin% names(tips)) {
    family_summ <- dplyr::mutate(
      dplyr::summarise(dplyr::group_by(tips, family),
        n_genus = dplyr::n_distinct(genus),
        n_spp = dplyr::n_distinct(species)
      ),
      genus = NA_character_
    )
    family_summ <- family_summ[!is.na(family_summ$family), ]
    genus_summ <- dplyr::left_join(
      dplyr::summarise(dplyr::group_by(tips, genus),
        n_genus = dplyr::n_distinct(genus),
        n_spp = dplyr::n_distinct(species)
      ),
      classification,
      by = "genus"
    )
    gf_summ <- dplyr::bind_rows(family_summ, genus_summ)
  } else {
    gf_summ <- dplyr::summarise(dplyr::group_by(tips, genus),
      n_genus = dplyr::n_distinct(genus),
      n_spp = dplyr::n_distinct(species)
    )
    gf_summ <- dplyr::mutate(gf_summ, family = NA)
  }

  gf_summ$grp <- 1:nrow(gf_summ)

  find_root <- function(xdf, tips, tree_df, show_warning) {
    target <- xdf$genus
    if (fam <- is.na(target)) target <- xdf$family
    sp_names <- if (fam) tips$species[tips$family == target] else tips$species[tips$genus == target]
    sp_names <- sp_names[!is.na(sp_names)]

    if (length(sp_names) == 1) {
      # Single species: its immediate parent is the basal (and root) node for the group.
      tip_row  <- tree_df[tree_df$label %fin% sp_names, ]
      root_row <- tree_df[tree_df$node == tip_row$parent, ]
      return(tibble::tibble(
        basal_node = root_row$label,
        root_node  = root_row$label,
        only_sp    = sp_names
      ))
    }

    # Multiple species: castor's C++ MRCA is ~10x faster than tidytree::MRCA on tbl_tree.
    # `tree` is the phylo object from the enclosing add_root_info() scope.
    mrca_num <- castor::get_mrca_of_set(tree, sp_names)
    mrca_row <- tree_df[tree_df$node == mrca_num, ]

    if (show_warning) {
      descts <- tidytree::offspring(tree_df, mrca_row$node, tiponly = TRUE)$label
      if (!setequal(sp_names, descts)) {
        message("Caution: Species in ", if (fam) "family" else "genus", " ", target,
                " do not form a monophyletic clade.")
      }
    }

    if (mrca_row$parent == mrca_row$node) {
      root_row <- mrca_row
    } else {
      root_row <- tree_df[tree_df$node == mrca_row$parent, ]
    }

    tibble::tibble(
      basal_node = mrca_row$label,
      root_node  = root_row$label,
      only_sp    = NA_character_
    )
  }

  # this takes time
  tree_df <- tidytree::as_tibble(tree)
  gf_summ2 <- dplyr::bind_rows(
    lapply(seq_len(nrow(gf_summ)), function(i)
      find_root(gf_summ[i, ], tips, tree_df, show_warning))
  )
  gf_summ2$grp <- seq_len(nrow(gf_summ))

  node_heights <- ape::branching.times(tree)

  gf_summ3 <- dplyr::left_join(gf_summ, dplyr::ungroup(gf_summ2), by = "grp")
  gf_summ3 <- dplyr::mutate(gf_summ3,
    root_time = node_heights[root_node],
    basal_time = node_heights[basal_node]
  )
  gf_summ3 <- gf_summ3[, c(
    "family", "genus", "basal_node", "basal_time",
    "root_node", "root_time", "n_genus", "n_spp", "only_sp"
  )]

  tree$genus_family_root <- gf_summ3

  tree
}

#' Remove trailing *
#'
#' @param tree A phylogeny generated by `get_tree(..., show_grafted = TRUE)` with trailing stars in tip labels.
#' @return A phylogeny after removing trailing stars.
#' @export
#'
rm_stars <- function(tree) {
  if (!inherits(tree, "phylo")) {
    stop("tree must have class of 'phylo'.")
  }
  tree$tip.label <- gsub("[*]*$", "", tree$tip.label)
  tree
}

# copied from plyr
progress_text <- function(style = 3, ...) {
  n <- 0
  txt <- NULL

  list(
    init = function(x) {
      txt <<- utils::txtProgressBar(max = x, style = style, ...)
      utils::setTxtProgressBar(txt, 0)
    },
    step = function() {
      n <<- n + 1
      utils::setTxtProgressBar(txt, n)
    },
    term = function() close(txt)
  )
}

progress_none <- function() {
  list(
    init = function(x) NULL,
    step = function() NULL,
    term = function() NULL
  )
}

create_progress_bar <- function(name = "text", ...) {
  if (!is.character(name)) {
    return(name)
  }
  name <- paste("progress", name, sep = "_")

  if (!exists(name, mode = "function")) {
    warning("Cannot find progress bar ", name, call. = FALSE)
    progress_none()
  } else {
    match.fun(name)(...)
  }
}


# Fast conversion from phylo to the 5-column list used by graft_all_cpp and
# bind_tip_core_cpp.  Avoids all tidytree/dplyr overhead of as_tibble.
phylo_to_vecs <- function(tree) {
  n_tips  <- ape::Ntip(tree)
  n_nodes <- ape::Nnode(tree)
  n_total <- n_tips + n_nodes

  # parent[i] = parent node number of node i (root is self-referential)
  par_vec <- integer(n_total)
  par_vec[tree$edge[, 2L]] <- tree$edge[, 1L]
  par_vec[n_tips + 1L] <- n_tips + 1L  # root (ape convention: always Ntip+1)

  bl_vec <- numeric(n_total)
  bl_vec[tree$edge[, 2L]] <- tree$edge.length

  node_labels <- if (!is.null(tree$node.label)) {
    tree$node.label
  } else {
    paste0("N", seq_len(n_nodes))
  }
  label_vec  <- c(tree$tip.label, node_labels)
  is_tip_vec <- c(rep(TRUE, n_tips), rep(FALSE, n_nodes))

  list(parent       = par_vec,
       node         = seq_len(n_total),
       branch.length = bl_vec,
       label        = label_vec,
       is_tip       = is_tip_vec)
}

## original contributed by Bradley Jones and modified by Guangchuang Yu
## copied from tidytree, modified to make it slightly faster
as_tree <- function(x) {
  x <- x[order(x$node), ] # tip.label must be in node-number order (ape convention)
  edge <- x[, c("parent", "node")]
  i <- which(edge[, 1] != 0 & edge[, 1] != edge[, 2])
  edge <- edge[i, ]
  if (is.null(x[["branch.length"]])) {
    edge.length <- NULL
  } else {
    edge.length <- x$branch.length[i]
  }
  tip.label <- as.character(x$label[x$is_tip])

  phylo <- list(
    edge = as.matrix(edge),
    edge.length = edge.length,
    tip.label = tip.label
  )

  node.label <- as.character(x$label[!x$is_tip])

  if (!all(is.na(node.label))) {
    phylo$node.label <- node.label
  }
  phylo$Nnode <- sum(!x[, "is_tip"])
  class(phylo) <- "phylo"
  return(phylo)
}

as_tree_isTip <- function(x) {
  # x = as.data.frame(x)
  i <- which(x$parent != 0 & x$parent != x$node)
  edge <- x[i, c("parent", "node")]
  if (is.null(x[["branch.length"]])) {
    edge.length <- NULL
  } else {
    edge.length <- x$branch.length[i]
  }
  tip.label <- x$label[x$isTip]

  phylo <- list(
    edge = edge,
    edge.length = edge.length,
    tip.label = tip.label
  )

  node.label <- x$label[!x$isTip]

  if (!all(is.na(node.label))) {
    phylo$node.label <- node.label
  }
  phylo$Nnode <- sum(!x[, "isTip"])
  class(phylo) <- "phylo"
  return(phylo)
}

#' Extract grafting status information as a data frame
#'
#' @param tree A phylogeny generated by `get_tree(...)` with trailing stars in tip labels.
#' @return A tibble with three columns: tip_label, species, and status.
#' @export
#'
get_graft_status <- function(tree) {
  tree$graft_status
}

context("Bug fix regression tests")

# Small user-provided trees used across multiple tests below.
# rosaceae_tree: ultrametric, 5 tips, 2 genera (Rosa, Rubus)
rosaceae_tree = ape::read.tree(text = paste0(
  "(((Rosa_setigera:3.4,Rosa_arkansana:3.4):2.0,",
  "Rosa_acicularis:5.4):5.0,",
  "(Rubus_odoratus:4.0,Rubus_parviflorus:4.0):6.4);"
))
rosaceae_tree$node.label = paste0("N", seq_len(ape::Nnode(rosaceae_tree)))

# tiny_tree: 3 tips across 2 families; Sabiaceae has only 1 member
tiny_tree = ape::read.tree(text = paste0(
  "((Rosa_setigera:3.4,Rosa_arkansana:3.4):5.0,",
  "Meliosma_oldenburgii:8.4);"
))
tiny_tree$node.label = paste0("N", seq_len(ape::Nnode(tiny_tree)))

# -----------------------------------------------------------------------
# Bug 2 (lines 186, 206, 208): close_sp for a genus not yet in the tree
#   Before: isTRUE() missing at line 186 caused logical(0) crash; root_time
#   stored branch.length instead of node age; idx_row not updated after
#   tibble::add_row() so n_spp stayed at 1 for subsequent congenerics.
# -----------------------------------------------------------------------

test_that("close_sp for new genus (at_basal_node) does not crash and produces valid tree", {
  sp = tibble::tibble(
    species     = c("Rosa_setigera", "Rosa_arkansana", "Rosa_acicularis",
                    "Rubus_odoratus", "Rubus_parviflorus", "Novus_sp1"),
    genus       = c("Rosa", "Rosa", "Rosa", "Rubus", "Rubus", "Novus"),
    family      = rep("Rosaceae", 6),
    close_sp    = c(NA, NA, NA, NA, NA, "Rosa_acicularis"),
    close_genus = rep(NA_character_, 6)
  )
  tt = suppressMessages(suppressWarnings(
    get_tree(sp, tree = rosaceae_tree, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  expect_s3_class(tt, "phylo")
  expect_equal(ape::Ntip(tt), nrow(sp))
  expect_false(any(tt$edge.length < -1e-8))
  expect_false(any(is.nan(tt$edge.length) | is.na(tt$edge.length)))
})

test_that("close_sp for new genus + random_below_basal (3 congenerics) produces valid tree", {
  # Novus_sp1 introduced via close_sp; Novus_sp2 and Novus_sp3 are then
  # grafted normally and exercise the root_time lookup on line 232.
  sp = tibble::tibble(
    species     = c("Rosa_setigera", "Rosa_arkansana", "Rosa_acicularis",
                    "Rubus_odoratus", "Rubus_parviflorus",
                    "Novus_sp1", "Novus_sp2", "Novus_sp3"),
    genus       = c("Rosa", "Rosa", "Rosa", "Rubus", "Rubus",
                    "Novus", "Novus", "Novus"),
    family      = rep("Rosaceae", 8),
    close_sp    = c(NA, NA, NA, NA, NA, "Rosa_acicularis", NA, NA),
    close_genus = rep(NA_character_, 8)
  )
  set.seed(1)
  tt = suppressMessages(suppressWarnings(
    get_tree(sp, tree = rosaceae_tree, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
  expect_s3_class(tt, "phylo")
  expect_equal(ape::Ntip(tt), nrow(sp))
  expect_false(any(tt$edge.length < -1e-8))
  expect_false(any(is.nan(tt$edge.length) | is.na(tt$edge.length)))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

# -----------------------------------------------------------------------
# Bug 3 (lines 260-261): when a family has exactly 1 tip and a new genus is
#   grafted there, root_node and root_time were set equal to basal_node and
#   basal_time. On random_below_basal for a 3rd congeneric, the difference
#   root_time - basal_time was 0, giving sum(bls) == 0 and division by zero.
# -----------------------------------------------------------------------

test_that("family with 1 member + new genus + random_below_basal does not divide by zero", {
  sp = tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana",
                "Meliosma_oldenburgii",
                "Sabia_sp1", "Sabia_sp2", "Sabia_sp3"),
    genus   = c("Rosa", "Rosa", "Meliosma", "Sabia", "Sabia", "Sabia"),
    family  = c("Rosaceae", "Rosaceae", "Sabiaceae",
                "Sabiaceae", "Sabiaceae", "Sabiaceae")
  )
  set.seed(2)
  tt = suppressMessages(suppressWarnings(
    get_tree(sp, tree = tiny_tree, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
  expect_s3_class(tt, "phylo")
  expect_equal(ape::Ntip(tt), nrow(sp))
  expect_false(any(tt$edge.length < -1e-8))
  expect_false(any(is.nan(tt$edge.length) | is.na(tt$edge.length)))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

test_that("family with 1 member + new genus + at_basal_node still works", {
  sp = tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana",
                "Meliosma_oldenburgii",
                "Sabia_sp1", "Sabia_sp2"),
    genus   = c("Rosa", "Rosa", "Meliosma", "Sabia", "Sabia"),
    family  = c("Rosaceae", "Rosaceae", "Sabiaceae", "Sabiaceae", "Sabiaceae")
  )
  tt = suppressMessages(suppressWarnings(
    get_tree(sp, tree = tiny_tree, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  expect_s3_class(tt, "phylo")
  expect_equal(ape::Ntip(tt), nrow(sp))
  expect_false(any(tt$edge.length < -1e-8))
})

# -----------------------------------------------------------------------
# Bug 5 (line 360): the ultrametric correction `branch.length - ages_diff`
#   could produce negative branch lengths when a tip is farther from the
#   root than the median. The fix wraps the correction in pmax(0, ...).
# -----------------------------------------------------------------------

test_that("ultrametric correction never produces negative branch lengths", {
  sp = tibble::tibble(
    species = c("Carya_alba", "Carya_cordiformis", "Meliosma_laui",
                "Acer_cordatum", "Fraxinus_mandshurica", "Ormosia_pinnata"),
    genus   = c("Carya", "Carya", "Meliosma", "Acer", "Fraxinus", "Ormosia"),
    family  = c("Juglandaceae", "Juglandaceae", "Sabiaceae",
                "Sapindaceae", "Oleaceae", "Fabaceae")
  )
  tt = suppressMessages(get_tree(sp, taxon = "plant", scenario = "at_basal_node"))
  expect_false(any(tt$edge.length < -1e-8))

  set.seed(42)
  tt_rbb = suppressMessages(get_tree(sp, taxon = "plant", scenario = "random_below_basal"))
  expect_false(any(tt_rbb$edge.length < -1e-8))
  expect_true(ape::is.ultrametric(tt_rbb, tol = 1e-4))
})

test_that("random_below_basal output is ultrametric when megatree is ultrametric", {
  set.seed(7)
  tt = suppressMessages(
    get_tree(test_plant_list, taxon = "plant", scenario = "random_below_basal")
  )
  expect_s3_class(tt, "phylo")
  expect_false(any(tt$edge.length < -1e-8))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

# -----------------------------------------------------------------------
# graft_status: returned tibble must have correct columns and cover all
# species regardless of grafting outcome.
# -----------------------------------------------------------------------

test_that("get_graft_status returns correct structure", {
  tt = suppressMessages(
    get_tree(test_plant_list, taxon = "plant", scenario = "at_basal_node")
  )
  gs = get_graft_status(tt)
  expect_s3_class(gs, "data.frame")
  expect_true(all(c("tip_label", "species", "status") %in% names(gs)))
  # all requested species must appear in graft_status
  expect_true(all(test_plant_list$species %in% gs$species))
  # status values must be from the known set
  valid_status = c("grafted at genus level", "grafted at family level",
                   "exisiting species in the megatree",
                   "skipped as no co-family in the megatree")
  expect_true(all(gs$status %in% valid_status))
})

context("Core functionality")

# -----------------------------------------------------------------------
# 1. Output species set: derived tree must contain exactly the requested
#    species -- no more, no fewer.
# -----------------------------------------------------------------------

test_that("output tip labels match requested species exactly", {
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant"))
  expect_setequal(tt$tip.label, test_plant_list$species)
})

test_that("output tree is rooted", {
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant"))
  expect_true(ape::is.rooted(tt))
})

test_that("output tip labels match for user-provided tree", {
  tt = suppressMessages(
    get_tree(test_tree_sp_df, tree = test_tree, taxon = "plant", tree_by_user = TRUE)
  )
  expected = gsub("\\*", "", tt$tip.label)      # rm_stars equivalent
  expect_setequal(expected, test_tree_sp_df$species)
})

# -----------------------------------------------------------------------
# 2. graft_status content: correct levels for known in/out species.
#    From helper_splists.R:
#      in-tree:  Carya_alba, Carya_cordiformis, Stewartia_sinensis,
#                Pterocarya_hupehensis, Ulmus_davidiana, Diplospora_dubia
#      out-tree: Meliosma_laui, Acer_cordatum, Fraxinus_mandshurica,
#                Ormosia_pinnata, Aglaia_dasyclada, Claoxylon_indicum
# -----------------------------------------------------------------------

test_that("graft_status reports 'exisiting species' for species present in megatree", {
  in_tree = c("Carya_alba", "Carya_cordiformis", "Stewartia_sinensis",
              "Pterocarya_hupehensis", "Ulmus_davidiana", "Diplospora_dubia")
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant"))
  gs = get_graft_status(tt)
  in_status = gs$status[gs$species %in% in_tree]
  expect_true(all(in_status == "exisiting species in the megatree"))
})

test_that("graft_status reports grafted status for species absent from megatree", {
  out_tree = c("Meliosma_laui", "Acer_cordatum", "Fraxinus_mandshurica",
               "Ormosia_pinnata", "Aglaia_dasyclada", "Claoxylon_indicum")
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant"))
  gs = get_graft_status(tt)
  out_status = gs$status[gs$species %in% out_tree]
  expect_true(all(out_status %in% c("grafted at genus level", "grafted at family level")))
})

test_that("graft_status covers every requested species", {
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant"))
  gs = get_graft_status(tt)
  expect_true(all(test_plant_list$species %in% gs$species))
})

# -----------------------------------------------------------------------
# 3. show_grafted and rm_stars
# -----------------------------------------------------------------------

test_that("show_grafted = TRUE adds asterisks to grafted tips only", {
  in_tree = c("Carya_alba", "Carya_cordiformis", "Stewartia_sinensis",
              "Pterocarya_hupehensis", "Ulmus_davidiana", "Diplospora_dubia")
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant", show_grafted = TRUE))

  # species in the megatree must have no asterisk suffix
  in_tips = tt$tip.label[gsub("\\*+$", "", tt$tip.label) %in% in_tree]
  expect_false(any(grepl("\\*$", in_tips)))

  # grafted species must carry one or two asterisks
  expect_true(any(grepl("\\*$", tt$tip.label)))
})

test_that("show_grafted = FALSE (default) produces no asterisks in tip labels", {
  tt = suppressMessages(get_tree(test_plant_list, taxon = "plant", show_grafted = FALSE))
  expect_false(any(grepl("\\*", tt$tip.label)))
})

test_that("rm_stars removes all trailing asterisks and preserves tip count", {
  tt       = suppressMessages(get_tree(test_plant_list, taxon = "plant", show_grafted = TRUE))
  tt_clean = rm_stars(tt)
  expect_false(any(grepl("\\*", tt_clean$tip.label)))
  expect_equal(ape::Ntip(tt_clean), ape::Ntip(tt))
  expect_setequal(tt_clean$tip.label, test_plant_list$species)
})

# -----------------------------------------------------------------------
# 4. add_root_info: structure and internal consistency
# -----------------------------------------------------------------------

test_that("add_root_info attaches genus_family_root with required columns", {
  cls = unique(rtrees::classifications[rtrees::classifications$taxon == "plant",
                                       c("genus", "family")])
  tt  = add_root_info(test_tree, classification = cls)
  gfr = tt$genus_family_root

  expect_s3_class(gfr, "data.frame")
  expect_true(all(c("family", "genus", "basal_node", "basal_time",
                    "root_node", "root_time", "n_genus", "n_spp", "only_sp") %in% names(gfr)))
})

test_that("add_root_info: all node labels in genus_family_root exist in the tree", {
  cls = unique(rtrees::classifications[rtrees::classifications$taxon == "plant",
                                       c("genus", "family")])
  tt  = add_root_info(test_tree, classification = cls)
  gfr = tt$genus_family_root

  tree_labels = c(tt$tip.label, tt$node.label)
  expect_true(all(gfr$basal_node %in% tree_labels))
  expect_true(all(gfr$root_node  %in% tree_labels))
})

test_that("add_root_info: root_time >= basal_time for every row", {
  cls = unique(rtrees::classifications[rtrees::classifications$taxon == "plant",
                                       c("genus", "family")])
  tt  = add_root_info(test_tree, classification = cls)
  gfr = tt$genus_family_root
  # invariant required by random_below_basal branch-length weighting
  expect_true(all(gfr$root_time >= gfr$basal_time - 1e-8))
})

test_that("add_root_info: n_spp matches observed tip counts per genus", {
  cls = unique(rtrees::classifications[rtrees::classifications$taxon == "plant",
                                       c("genus", "family")])
  tt  = add_root_info(test_tree, classification = cls)
  gfr = tt$genus_family_root

  genus_rows = gfr[!is.na(gfr$genus), ]
  for (g in genus_rows$genus) {
    actual   = sum(grepl(paste0("^", g, "_"), tt$tip.label))
    recorded = genus_rows$n_spp[genus_rows$genus == g]
    expect_equal(actual, recorded, info = paste("n_spp mismatch for genus", g))
  }
})

# -----------------------------------------------------------------------
# 5. bind_tip: branch length correctness for different frac values.
#    Uses a small tree with known branch lengths so expectations are exact.
#    Tree:  ((A:10,B:10):5,C:15);  → root age 15, internal node age 10
# -----------------------------------------------------------------------

known_tree = ape::read.tree(text = "((A:10,B:10):5,C:15);")
known_tree$node.label = paste0("N", seq_len(ape::Nnode(known_tree)))

test_that("bind_tip to a tip with frac = 0.5 splits branch correctly", {
  # A has branch.length = 10; new node inserted at height 5 (= 10 * 0.5)
  # both children of new node should have branch.length = 5
  tt2     = bind_tip(known_tree, where = "A", tip_label = "New_sp", frac = 0.5)
  tt2_tbl = tidytree::as_tibble(tt2)

  target_bl  = tt2_tbl$branch.length[tt2_tbl$label == "A"]
  new_tip_bl = tt2_tbl$branch.length[tt2_tbl$label == "New_sp"]

  expect_equal(target_bl,  5.0, tolerance = 1e-8)
  expect_equal(new_tip_bl, 5.0, tolerance = 1e-8)
})

test_that("bind_tip to a tip with frac = 0.25 puts new node closer to parent", {
  # A has branch.length = 10; frac = 0.25 means new node 2.5 below parent
  # both children get branch.length = 10 * 0.75 = 7.5
  tt2     = bind_tip(known_tree, where = "A", tip_label = "New_sp", frac = 0.25)
  tt2_tbl = tidytree::as_tibble(tt2)

  target_bl  = tt2_tbl$branch.length[tt2_tbl$label == "A"]
  new_tip_bl = tt2_tbl$branch.length[tt2_tbl$label == "New_sp"]

  expect_equal(target_bl,  7.5, tolerance = 1e-8)
  expect_equal(new_tip_bl, 7.5, tolerance = 1e-8)
})

test_that("bind_tip to an internal node attaches tip with branch length = node age", {
  # internal node N1 is at height 10; new tip attached to it should have bl = 10
  node_lbl = "N1"
  node_hts = ape::branching.times(known_tree)

  tt2     = bind_tip(known_tree, where = node_lbl, tip_label = "New_sp")
  tt2_tbl = tidytree::as_tibble(tt2)

  new_tip_bl = tt2_tbl$branch.length[tt2_tbl$label == "New_sp"]
  expect_equal(new_tip_bl, unname(node_hts[node_lbl]), tolerance = 1e-8)
})

# -----------------------------------------------------------------------
# 6. sp_list_df: content correctness
# -----------------------------------------------------------------------

test_that("sp_list_df extracts genus correctly from species name", {
  df = sp_list_df(c("Quercus_robur", "Pinus_sylvestris", "Betula_pendula"))
  expect_equal(df$genus,   c("Quercus", "Pinus", "Betula"))
  expect_equal(df$species, c("Quercus_robur", "Pinus_sylvestris", "Betula_pendula"))
})

test_that("sp_list_df normalises spaces to underscores and capitalises first letter", {
  df = sp_list_df(c("quercus robur", "pinus sylvestris"))
  expect_equal(df$species, c("Quercus_robur", "Pinus_sylvestris"))
  expect_equal(df$genus,   c("Quercus", "Pinus"))
})

test_that("sp_list_df handles phylomatic format (family/genus/species)", {
  phy = c("juglandaceae/Carya/Carya_alba",
          "sapindaceae/Acer/Acer_cordatum",
          "oleaceae/Fraxinus/Fraxinus_mandshurica")
  df = sp_list_df(phy)
  expect_equal(nrow(df), 3)
  expect_true("family" %in% names(df))
  expect_equal(df$species, c("Carya_alba", "Acer_cordatum", "Fraxinus_mandshurica"))
  expect_equal(df$genus,   c("Carya", "Acer", "Fraxinus"))
  expect_equal(tolower(df$family), c("juglandaceae", "sapindaceae", "oleaceae"))
})

test_that("sp_list_df de-duplicates repeated species names", {
  df = sp_list_df(c("Quercus_robur", "Quercus_robur", "Pinus_sylvestris"))
  expect_equal(nrow(df), 2)
})

test_that("sp_list_df maps genus to family correctly for plants", {
  df = sp_list_df(c("Carya_alba", "Quercus_robur"), taxon = "plant")
  expect_true("family" %in% names(df))
  expect_equal(df$family[df$genus == "Carya"],  "Juglandaceae")
  expect_equal(df$family[df$genus == "Quercus"], "Fagaceae")
})

# -----------------------------------------------------------------------
# 7. multiPhylo output: all trees valid with identical tip sets
# -----------------------------------------------------------------------

test_that("bird multiPhylo output has correct length and all trees are phylo", {
  tt = suppressMessages(get_tree(test_bird_list, taxon = "bird"))
  expect_true(inherits(tt, c("multiPhylo", "list")))
  expect_true(all(sapply(tt, inherits, "phylo")))
})

test_that("all trees in bird multiPhylo have correct tip count", {
  tt = suppressMessages(get_tree(test_bird_list, taxon = "bird"))
  expect_true(all(sapply(tt, ape::Ntip) == nrow(test_bird_list)))
})

test_that("all trees in bird multiPhylo share the same tip label set", {
  tt    = suppressMessages(get_tree(test_bird_list, taxon = "bird"))
  tips1 = sort(tt[[1]]$tip.label)
  same  = sapply(tt, function(x) identical(sort(x$tip.label), tips1))
  expect_true(all(same))
})

test_that("get_graft_status works on individual trees from multiPhylo output", {
  tt = suppressMessages(get_tree(test_bird_list, taxon = "bird", show_grafted = TRUE))
  gs = get_graft_status(tt[[1]])
  expect_s3_class(gs, "data.frame")
  expect_true(all(c("tip_label", "species", "status") %in% names(gs)))
  expect_true(all(test_bird_list$species %in% gs$species))
})

context("Testing utils")

test_that("Convert a vector of sp to data frame", {
  expect_equal(ncol(sp_list_df(test_plant_list$species, taxon = "plant")), 3)
  expect_equal(ncol(sp_list_df(test_fish_list$species, taxon = "fish")), 3)
  
  expect_warning(sp_list_df(test_plant_list$species, taxon = "fish"))
  expect_warning(sp_list_df(test_fish_list$species, taxon = "plant"))
})

test_that("Working with a data frame", {
  expect_equal(ncol(sp_list_df(test_plant_list[, c("species"), drop = FALSE])), 2)
  expect_equal(ncol(sp_list_df(test_plant_list[, c("species"), drop = FALSE], taxon = "plant")), 3)
  expect_equal(ncol(sp_list_df(test_plant_list[, c("species", "genus"), drop = FALSE])), 2)
})

test_that("Convert the first letter to upper case", {
  expect_equal(cap_first_letter("aha"), "Aha")
  expect_equal(cap_first_letter("Aha"), "Aha")
})

#' Fetching phylogenies from mega-trees
#'
#' @name rtrees-package
#' @aliases rtrees
#' @docType package
#' @author Daijiang Li <daijianglee@@gmail.com>
#' @keywords internal
#' @useDynLib rtrees, .registration = TRUE
#' @importFrom Rcpp evalCpp
"_PACKAGE"

#' Get one or multiple trees from megatree(s)
#'
#' For some taxa groups, there are multiple posterior megatrees. It is a common
#' task to derive a phylogeny from each of these (or a random subset of) megatrees.
#'
#'
#'
#' Derive a phylogeny from a mega-tree
#'
#' For a list of species, generate a phylogeny or multiple phylogenies from a
#' provided mega-tree or mega-trees. If a species is
#' not in the mega-tree, it will be grafted to the mega-tree with two scenarios.
#'
#' @param sp_list A character vector or a data frame with at least three columns: species, genus, family. Species column
#' holds the species for which we want to have a phylogeny. It can also have two optional columns:
#' close_sp and close_genus. We can specify the closest species/genus of the species based on
#' expert knowledge. If specified, the new species will be grafted to that particular location.
#'
#' It can also be a string vector if `taxon` is specified. Though it probably is a better idea
#' to prepare your data frame with [sp_list_df()]. The string vector can also have the same format
#' as that required by Phylomatic (i.e., family/genus/genus_sp).
#' @param tree A mega-tree with class `phylo` or a list of mega-trees with class `multiPhylo`.
#'  Optional if `taxon` is specified, in which case, a default
#' mega-phylogeny (or a set of 100 randomly selected posterior phylogenies) will be used
#' (see their own documentations from the `megatrees` package).
#'
#' - For amphibian, the mega-trees are loaded via [megatrees::get_tree_amphibian_n100()].
#' - For bee, the mega-tree is [megatrees::tree_bee], with the bootstrap option loaded via [megatrees::get_tree_bee_n100()].
#' - For butterfly, the mega-tee is [megatrees::tree_butterfly].
#' - For bird, the mega-trees are loaded via [megatrees::get_tree_bird_n100()].
#' - For fish, the mega-tree is [megatrees::tree_fish_12k], with the all-taxon option loaded via [megatrees::get_tree_fish_32k_n50()].
#' - For mammal, the default mega-trees are loaded via [megatrees::get_tree_mammal_n100_vertlife()], with [megatrees::get_tree_mammal_n100_phylacine()] be the other option.
#' - For plant, the default mega-tree is [megatrees::tree_plant_otl] (Smith and Brown 2018);
#'   [megatrees::tree_plant_Carruthers] (Carruthers et al. 2026) is also available; 100 posterior trees are loaded via
#'   [megatrees::get_tree_plant_n100_Carruthers()].
#' - For reptile, the mega-trees are loaded via [megatrees::get_tree_reptile_n100()].
#' - For shark, ray, and chimaeras, the mega-trees are loaded via [megatrees::get_tree_shark_ray_n100()].
#'
#' @param taxon The taxon of species in the `sp_list`. Currently, can be `amphibian`, `bird`, `fish`, `mammal`, `plant`, `reptile`, or `shark_ray`.
#' @param scenario How to insert a species into the mega-tree?
#' - In both scenarioes, if there is only 1 species in the genus or family, a new node will be inserted to
#' the middle point of this only species' branch length and the new species will be attached to this new
#' node.
#' - If `scenario = "at_basal_node"`, a species is attached to the basal node of the same genus or the same family
#' if the mega-tree does not have any species of this genus.
#' - If `scenario = "random_below_basal"`, a species is attached to a randomly selected node that is at or below the
#' basal node of the same genus of the same family if the mega-tree does not have any species in this genus.
#' The probability of node been selected is proportional to its branch length.
#' Because of the random sampling involved, you may want to run several times to get a collection of
#' derived phylogenies.
#' @param show_grafted Whether to indicate which species was grafted onto the mega-tree.
#' If `TRUE`, a `*` will be appended to the species name on the tip if it was grafted within
#' the same genus; `**` will be appended if it was grafted within the same family.
#' @param tree_by_user Is the mega-tree provided by user? Default is `FALSE` but it will be automatically set to `TRUE` when the class of `tree` is
#' `multiPhylo` since we don't provide any such mega-trees here.
#' @param mc_cores Number of cores to parallel processing when `tree` is a list of large number of trees. The default is the number of available cores minus 2.
#' @param .progress Form of progress bar, default to be text.
#' @param fish_tree Which fish tree do you want to use? If it is "timetree" (default), it will be the smaller time tree with 11638 species
#' that all have sequence data; if it is "all-taxon", then it will be the 100 larger posterior phylogenies with 31516 soecues.
#' @param mammal_tree Which set of mammal trees to use? If it is "vertlife" (default), then 100 randomly selected posterior phylogenies provided
#' by Vertlife will be used; if it is "phylacine", then 100 randomly selected posterior phylogenies provided by PHYLACINE will be used.
#' @param bee_tree Which bee tree to use? If it is "maximum-likelihood" (default), the a single maximum likelihood tree will be used. If it is "bootstrap", then a set of 100 randomly selected posterior phylogenies will be used. All trees are provided by the [Bee Tree of Life](http://beetreeoflife.org).
#' @param plant_tree Which plant tree to use? If `"tree_plant_otl"` (default), the Smith and Brown (2018) tree
#'   is used. If `"tree_plant_Carruthers"`, the single best-scoring tree from Carruthers et al. (2026) is used.
#'   If `"tree_plant_n100_Carruthers"`, 100 randomly selected posterior phylogenies from Carruthers et al. (2026)
#'   are downloaded and used (large file, ~135 MB, cached after first use).
#' @param dt Whether to use data.table version to bind tips [bind_tip]. The default is `TRUE` as it maybe slightly faster.
#' @return A phylogeny for the species required, with class `phylo`;
#' or a list of phylogenies with class `multiPhylo` depends on the input `tree`. Within each phylogeny, the grafted status of all species was saved as a data frame named as "graft_status".
#' @export
#' @examples
#' test_sp <- c(
#'   "Serrasalmus_geryi", "Careproctus_reinhardti", "Gobiomorphus_coxii",
#'   "Periophthalmus_barbarus", "Prognichthys_glaphyrae", "Barathronus_bicolor",
#'   "Knipowitschia_croatica", "Rhamphochromis_lucius", "Neolissochilus_tweediei",
#'   "Haplochromis_nyanzae", "Astronesthes_micropogon", "Sanopus_reticulatus"
#' )
#' test_tree <- get_tree(
#'   sp_list = test_sp,
#'   taxon = "fish",
#'   show_grafted = TRUE
#' )
get_tree <- function(sp_list, tree, taxon = NULL,
                     # multiple_trees = TRUE,
                     scenario = c("at_basal_node", "random_below_basal"),
                     show_grafted = FALSE,
                     tree_by_user = FALSE,
                     mc_cores = future::availableCores() - 2, .progress = "text",
                     fish_tree = c("timetree", "all-taxon"),
                     mammal_tree = c("vertlife", "phylacine"),
                     bee_tree = c("maximum-likelihood", "bootstrap"),
                     plant_tree = c("tree_plant_otl", "tree_plant_Carruthers", "tree_plant_n100_Carruthers"),
                     dt = TRUE) {
  scenario <- match.arg(scenario)

  if (missing(tree) & is.null(taxon)) {
    stop("Please specify at least a tree or a taxon group.")
  }
  if (missing(tree) & !is.null(taxon)) { # pick default tree
    if (taxon == "plant") {
      plant_tree <- match.arg(plant_tree)
      if (plant_tree == "tree_plant_otl") tree <- megatrees::tree_plant_otl
      if (plant_tree == "tree_plant_Carruthers") tree <- megatrees::tree_plant_Carruthers
      if (plant_tree == "tree_plant_n100_Carruthers") tree <- megatrees::get_tree_plant_n100_Carruthers()
    }
    if (taxon == "bird") tree <- megatrees::get_tree_bird_n100()
    if (taxon == "butterfly") {
      warning("Classification of butterfly is not complete; it will work better if you can prepare the family information for your species.")
      tree <- megatrees::tree_butterfly
    }
    if (taxon == "amphibian") tree <- megatrees::get_tree_amphibian_n100()
    if (taxon == "reptile") tree <- megatrees::get_tree_reptile_n100()
    if (taxon == "shark_ray") tree <- megatrees::get_tree_shark_ray_n100()
    if (taxon == "mammal") {
      mammal_tree <- match.arg(mammal_tree)
      if (mammal_tree == "vertlife") tree <- megatrees::get_tree_mammal_n100_vertlife()
      if (mammal_tree == "phylacine") tree <- megatrees::get_tree_mammal_n100_phylacine()
    }
    if (taxon == "fish") {
      fish_tree <- match.arg(fish_tree)
      if (fish_tree == "timetree") tree <- megatrees::tree_fish_12k
      if (fish_tree == "all-taxon") tree <- megatrees::get_tree_fish_32k_n50()
    }
    if (taxon == "bee") {
      bee_tree <- match.arg(bee_tree)
      if (bee_tree == "maximum-likelihood") tree <- megatrees::tree_bee
      if (bee_tree == "bootstrap") tree <- megatrees::get_tree_bee_n100()
    }
  }

  if (inherits(tree, "phylo")) { # one phylo
    return(get_one_tree(
      sp_list = sp_list, tree = tree, taxon = taxon,
      scenario = scenario, show_grafted = show_grafted,
      tree_by_user = tree_by_user, .progress = .progress,
      dt = dt
    ))
  }

  if ((inherits(tree, "multiPhylo") | inherits(tree, "list")) &
    length(tree) > 1 &
    inherits(tree[[1]], "phylo") # a list of multiple phylo
  ) {
    # if(.Platform$OS.type == "windows" & mc_cores > 1){
    #   warning("parallel does not work on Windows")
    # }

    if (is.null(tree[[1]]$genus_family_root)) tree_by_user <- TRUE

    if (mc_cores > 1) {
      if (mc_cores > future::availableCores()) {
        message("The specified mc_cores is larger than what available")
        mc_cores <- future::availableCores() - 1
      }
      future::plan(future::multisession, workers = mc_cores)
      # run it once just to get the messages, not ideal but I did not find another way
      # to only show the message once yet.
      out0 <- rtrees::get_one_tree(sp_list,
        tree = tree[[1]], taxon = taxon,
        scenario = scenario, show_grafted = show_grafted,
        tree_by_user = tree_by_user,
        .progress = "none", dt = dt
      )

      out <- furrr::future_map(tree[-1], function(i) {
        suppressMessages(rtrees::get_one_tree(sp_list,
          tree = i, taxon = taxon,
          scenario = scenario, show_grafted = show_grafted,
          tree_by_user = tree_by_user,
          .progress = "none", dt = dt
        ))
        # .progress = "none" # hide progress bar
      }, .progress = TRUE, .options = furrr::furrr_options(seed = TRUE))

      # put the first generated phylogeny back and in order
      out[[length(tree)]] <- out0
      names(out)[length(tree)] <- names(tree)[1]
      out <- c(out[length(tree)], out[-length(tree)])
    } else {
      ## TO DO: test .progress issue with lapply()
      out <- lapply(tree, get_one_tree,
        sp_list = sp_list, taxon = taxon,
        scenario = scenario, show_grafted = show_grafted,
        tree_by_user = tree_by_user, .progress = .progress, dt = dt
      )
    }

    class(out) <- "multiPhylo"
    return(out)
  }
}

# Extracted from test_invariants.R:242

# prequel ----------------------------------------------------------------------
context("Algorithmic invariants")
inv_base <- ape::read.tree(text = paste0(
  "(((Rosa_setigera:3.4,Rosa_arkansana:3.4):2.0,",
  "Rosa_acicularis:5.4):5.0,",
  "(Rubus_odoratus:4.0,Rubus_parviflorus:4.0):6.4);"
))
inv_base$node.label <- paste0("N", seq_len(ape::Nnode(inv_base)))
inv_base_coph <- ape::cophenetic.phylo(inv_base)
inv_tiny <- ape::read.tree(text = paste0(
  "((Rosa_setigera:3.4,Rosa_arkansana:3.4):5.0,",
  "Meliosma_oldenburgii:8.4);"
))
inv_tiny$node.label <- paste0("N", seq_len(ape::Nnode(inv_tiny)))
is_pure_clade <- function(tree, tips) {
  if (length(tips) < 2L) return(NA)
  mrca <- ape::getMRCA(tree, tips)
  if (is.null(mrca)) return(FALSE)
  if (mrca == ape::Ntip(tree) + 1L) return(setequal(tree$tip.label, tips))
  setequal(ape::extract.clade(tree, node = mrca)$tip.label, tips)
}

# test -------------------------------------------------------------------------
sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana",
                "Novus_sp1", "Novus_sp2", "Novus_sp3",
                "Alter_sp1", "Alter_sp2"),
    genus   = c("Rosa", "Rosa", "Novus", "Novus", "Novus", "Alter", "Alter"),
    family  = rep("Rosaceae", 7)
  )
set.seed(7)
tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
novus_tips <- grep("^Novus_", tt$tip.label, value = TRUE)
alter_tips <- grep("^Alter_", tt$tip.label, value = TRUE)
expect_true(is_pure_clade(tt, novus_tips))
expect_true(is_pure_clade(tt, alter_tips))

context("Algorithmic invariants")

# ---------------------------------------------------------------------------
# Shared fixtures
# ---------------------------------------------------------------------------

# ultrametric base: Rosa (3 spp) + Rubus (2 spp), all Rosaceae
inv_base <- ape::read.tree(text = paste0(
  "(((Rosa_setigera:3.4,Rosa_arkansana:3.4):2.0,",
  "Rosa_acicularis:5.4):5.0,",
  "(Rubus_odoratus:4.0,Rubus_parviflorus:4.0):6.4);"
))
inv_base$node.label <- paste0("N", seq_len(ape::Nnode(inv_base)))
inv_base_coph <- ape::cophenetic.phylo(inv_base)

# tiny base: Rosa (2 spp) + Meliosma (1 sp), Rosaceae + Sabiaceae
inv_tiny <- ape::read.tree(text = paste0(
  "((Rosa_setigera:3.4,Rosa_arkansana:3.4):5.0,",
  "Meliosma_oldenburgii:8.4);"
))
inv_tiny$node.label <- paste0("N", seq_len(ape::Nnode(inv_tiny)))

# Check that `tips` form a pure clade: their MRCA subtree contains no other tips.
is_pure_clade <- function(tree, tips) {
  if (length(tips) < 2L) return(NA)
  mrca <- ape::getMRCA(tree, tips)
  if (is.null(mrca)) return(FALSE)
  if (mrca == ape::Ntip(tree) + 1L) return(setequal(tree$tip.label, tips))
  setequal(ape::extract.clade(tree, node = mrca)$tip.label, tips)
}

# -----------------------------------------------------------------------
# 1. Non-disturbance: grafting new species must not alter pairwise distances
#    of species already present in the base tree.
# -----------------------------------------------------------------------

test_that("at_basal_node: existing species distances unchanged after grafting", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Rubus_odoratus", "Rubus_parviflorus",
                "Novus_sp1", "Novus_sp2"),
    genus   = c("Rosa", "Rosa", "Rubus", "Rubus", "Novus", "Novus"),
    family  = rep("Rosaceae", 6)
  )
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  D <- ape::cophenetic.phylo(tt)
  expect_equal(D["Rosa_setigera",  "Rosa_arkansana"],
               inv_base_coph["Rosa_setigera", "Rosa_arkansana"],    tolerance = 1e-8)
  expect_equal(D["Rubus_odoratus", "Rubus_parviflorus"],
               inv_base_coph["Rubus_odoratus", "Rubus_parviflorus"], tolerance = 1e-8)
})

test_that("random_below_basal: existing species distances unchanged after grafting", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Rubus_odoratus", "Rubus_parviflorus",
                "Novus_sp1", "Novus_sp2"),
    genus   = c("Rosa", "Rosa", "Rubus", "Rubus", "Novus", "Novus"),
    family  = rep("Rosaceae", 6)
  )
  set.seed(42)
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
  D <- ape::cophenetic.phylo(tt)
  expect_equal(D["Rosa_setigera",  "Rosa_arkansana"],
               inv_base_coph["Rosa_setigera", "Rosa_arkansana"],    tolerance = 1e-8)
  expect_equal(D["Rubus_odoratus", "Rubus_parviflorus"],
               inv_base_coph["Rubus_odoratus", "Rubus_parviflorus"], tolerance = 1e-8)
})

# -----------------------------------------------------------------------
# 2. Determinism: at_basal_node must produce bit-identical output for the
#    same input (it is a deterministic algorithm with no random choices).
# -----------------------------------------------------------------------

test_that("at_basal_node: repeated calls with same input produce identical trees", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Rubus_odoratus",
                "Novus_sp1", "Novus_sp2", "Novus_sp3"),
    genus   = c("Rosa", "Rosa", "Rubus", "Novus", "Novus", "Novus"),
    family  = rep("Rosaceae", 6)
  )
  tt1 <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  tt2 <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  sps <- sort(tt1$tip.label)
  D1  <- ape::cophenetic.phylo(tt1)[sps, sps]
  D2  <- ape::cophenetic.phylo(tt2)[sps, sps]
  expect_equal(max(abs(D1 - D2)), 0, tolerance = 1e-8)
})

# -----------------------------------------------------------------------
# 3. Stranded genus clade purity: multiple species from a genus absent from
#    the base tree should form a pure clade with no interlopers.
# -----------------------------------------------------------------------

test_that("at_basal_node: stranded genus species form a pure clade", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Rubus_odoratus",
                "Novus_sp1", "Novus_sp2", "Novus_sp3"),
    genus   = c("Rosa", "Rosa", "Rubus", "Novus", "Novus", "Novus"),
    family  = rep("Rosaceae", 6)
  )
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  novus_tips <- grep("^Novus_", tt$tip.label, value = TRUE)
  expect_length(novus_tips, 3)
  expect_true(is_pure_clade(tt, novus_tips))
})

test_that("random_below_basal: stranded genus species form a pure clade", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Rubus_odoratus",
                "Novus_sp1", "Novus_sp2", "Novus_sp3"),
    genus   = c("Rosa", "Rosa", "Rubus", "Novus", "Novus", "Novus"),
    family  = rep("Rosaceae", 6)
  )
  set.seed(1)
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
  novus_tips <- grep("^Novus_", tt$tip.label, value = TRUE)
  expect_length(novus_tips, 3)
  expect_true(is_pure_clade(tt, novus_tips))
})

# -----------------------------------------------------------------------
# 4. Star topology: stranded genus species (at_basal_node) all attach to
#    the same internal node, so their pairwise distances must all be equal.
# -----------------------------------------------------------------------

test_that("at_basal_node: stranded genus forms a star (all pairwise distances equal)", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Rubus_odoratus",
                "Novus_sp1", "Novus_sp2", "Novus_sp3"),
    genus   = c("Rosa", "Rosa", "Rubus", "Novus", "Novus", "Novus"),
    family  = rep("Rosaceae", 6)
  )
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  novus_tips <- grep("^Novus_", tt$tip.label, value = TRUE)
  D <- ape::cophenetic.phylo(ape::keep.tip(tt, novus_tips))
  dvals <- D[lower.tri(D)]
  expect_lt(diff(range(dvals)), 1e-6)
})

# -----------------------------------------------------------------------
# 5. n_spp == 1 family case: when the family has exactly one species in the
#    base tree, multiple new-genus species should still form a pure clade.
# -----------------------------------------------------------------------

test_that("at_basal_node: new genus in 1-species family forms a pure clade", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Meliosma_oldenburgii",
                "Sabia_sp1", "Sabia_sp2", "Sabia_sp3"),
    genus   = c("Rosa", "Rosa", "Meliosma", "Sabia", "Sabia", "Sabia"),
    family  = c("Rosaceae", "Rosaceae", "Sabiaceae",
                "Sabiaceae", "Sabiaceae", "Sabiaceae")
  )
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_tiny, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  sabia_tips <- grep("^Sabia_", tt$tip.label, value = TRUE)
  expect_length(sabia_tips, 3)
  expect_true(is_pure_clade(tt, sabia_tips))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

test_that("random_below_basal: new genus in 1-species family forms a pure clade", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana", "Meliosma_oldenburgii",
                "Sabia_sp1", "Sabia_sp2", "Sabia_sp3"),
    genus   = c("Rosa", "Rosa", "Meliosma", "Sabia", "Sabia", "Sabia"),
    family  = c("Rosaceae", "Rosaceae", "Sabiaceae",
                "Sabiaceae", "Sabiaceae", "Sabiaceae")
  )
  set.seed(3)
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_tiny, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
  sabia_tips <- grep("^Sabia_", tt$tip.label, value = TRUE)
  expect_length(sabia_tips, 3)
  expect_true(is_pure_clade(tt, sabia_tips))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

# -----------------------------------------------------------------------
# 6. Two stranded genera from the same family must each form their own
#    separate pure clade (no cross-contamination between genera).
# -----------------------------------------------------------------------

test_that("at_basal_node: two stranded genera form independent pure clades", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana",
                "Novus_sp1", "Novus_sp2", "Novus_sp3",
                "Alter_sp1", "Alter_sp2"),
    genus   = c("Rosa", "Rosa", "Novus", "Novus", "Novus", "Alter", "Alter"),
    family  = rep("Rosaceae", 7)
  )
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "at_basal_node", tree_by_user = TRUE)
  ))
  novus_tips <- grep("^Novus_", tt$tip.label, value = TRUE)
  alter_tips <- grep("^Alter_", tt$tip.label, value = TRUE)
  expect_true(is_pure_clade(tt, novus_tips))
  expect_true(is_pure_clade(tt, alter_tips))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

test_that("random_below_basal: two stranded genera form independent pure clades", {
  sp <- tibble::tibble(
    species = c("Rosa_setigera", "Rosa_arkansana",
                "Novus_sp1", "Novus_sp2", "Novus_sp3",
                "Alter_sp1", "Alter_sp2"),
    genus   = c("Rosa", "Rosa", "Novus", "Novus", "Novus", "Alter", "Alter"),
    family  = rep("Rosaceae", 7)
  )
  set.seed(7)
  tt <- suppressMessages(suppressWarnings(
    get_tree(sp, tree = inv_base, taxon = "plant",
             scenario = "random_below_basal", tree_by_user = TRUE)
  ))
  novus_tips <- grep("^Novus_", tt$tip.label, value = TRUE)
  alter_tips <- grep("^Alter_", tt$tip.label, value = TRUE)
  expect_true(is_pure_clade(tt, novus_tips))
  expect_true(is_pure_clade(tt, alter_tips))
  expect_true(ape::is.ultrametric(tt, tol = 1e-4))
})

#' Classifications of species
#' 
#' @description 
#' Genus and family information of different groups of taxon. 
#' 
#' - Plant classification information. Its sources include:
#'         + based on [`V.PhyloMaker::nodes.info.1`](https://github.com/jinyizju/V.PhyloMaker/tree/master/data)
#'         + based on The Plant List
#'         + [taxonlookup](https://github.com/traitecoevo/taxonlookup)
#'         + [Plants of the World online](http://plantsoftheworldonline.org/)
#'         
#' - Fish classification information was based on FishBase. There are 4,825 genus in this file.
#' <https://fishtreeoflife.org/downloads/PFC_taxonomy.csv.xz>
#' 
#' - Bee classification information was from [Bee Tree of Life](http://beetreeoflife.org/nomenclature/). Note that we used 'Subfamily' in their nomenclature file as "family" here. If a genus' Subfamily is missing, we used its Family.
#' 
#' - Bird classification information was based on BirdLife, which resulted in 2,391 genus.
#' <http://datazone.birdlife.org/species/taxonomy> However, based on the 
#' [taxonomy file](https://data.vertlife.org/birdtree/BLIOCPhyloMasterTax.csv)
#' of the Jetz et al. 2012 phylogeny, there are additional 117 genus that are not in the file
#' of BirdLife. Both are combined here, which leads to 2,508 genus.
#' 
#' - Mammal classification information was based on PHYLACINE, which has 1,400 genus. 
#' <https://github.com/MegaPast2Future/PHYLACINE_1.2/blob/master/Data/Taxonomy/Synonymy_table_valid_species_only.csv> Additional 
#' genus from Vertlife were added too. For the same genus from both PHYLACINE and Vertlife that have different family information,
#' I used the family from Vertlife as I found that they are mostly more accurate.
#' 
#' - Amphibian classification information was from [VertLife](https://data.vertlife.org/amphibiantree/download/amph_shl_new_Classification.csv).
#' 
#' - Reptile classification information was largely from [wikipedia](https://en.wikipedia.org/wiki/List_of_reptile_genera).
#' 
#' - Shark and Ray classification information was largely from NCBI.
#' 
#' - Butterfly classification information was from Kawahara et al. 2023, using the tip labels of their phylogeny.
#' 
#' @format A data frame with three columns: genus, family, and taxon (`plant`, `fish`, `bird`, `mammal`, `amphibian`, `reptile`, `shark_ray`, `bee`, `butterfly`). 
#' 
"classifications"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

bind_tip_core_cpp <- function(parent, node, branch_length, label, is_tip, where, tip_label, node_label_str, frac, new_node_above, node_height) {
    .Call(`_rtrees_bind_tip_core_cpp`, parent, node, branch_length, label, is_tip, where, tip_label, node_label_str, frac, new_node_above, node_height)
}

graft_all_cpp <- function(parent0, node0, bl0, label0, is_tip0, where_labels, new_tip_labels, new_node_labels, fracs, above_flags, node_heights_vec) {
    .Call(`_rtrees_graft_all_cpp`, parent0, node0, bl0, label0, is_tip0, where_labels, new_tip_labels, new_node_labels, fracs, above_flags, node_heights_vec)
}

fast_internal_offspring_cpp <- function(parent, node, is_tip, root_node) {
    .Call(`_rtrees_fast_internal_offspring_cpp`, parent, node, is_tip, root_node)
}

---
title: "Speed comparisons"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Speed comparisons}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Here, I compared the performance of `rtrees` (v1.0.0) and `V.PhyloMaker` (v0.1.0) (`V.PhyloMaker2` has almost the same speed as `V.PhyloMaker`, so no comparisons here). In most cases, `rtrees` is at least two times faster than `V.PhyloMaker` (Fig. below). For example, with 50 missing species to bind, the average time used by `rtrees` is 0.914 second while `V.PhyloMaker` took 13.6 seconds on average; with 5,000 missing species to bind, `rtrees` used 80.9 seconds on average while `V.PhyloMaker` used 187 seconds. In general, the time used by `rtrees` and `V.PhyloMaker` increased 15.6 and 34.8 seconds on average with 1,000 more missing species to be grafted, respectively. All tests were conducted on a 14' Macbook Pro with M1 pro chip.

Some may not think it is a big difference. However, for me as a user, waiting for more than 10 seconds to have about 50 species grafted is not ideal. Also the accumulated time can be large if one is doing this multiple times (e.g., doing some simulations that require thousands of iterations).

The R code used for the speed tests can be found [here](https://github.com/daijiang/rtrees_ms/blob/main/Rcode/speed_test.R).


```{r, out.width='90%', fig.width=7, fig.height=5}
library(ggplot2)
speed_out = readRDS(system.file("extdata", "rtrees_speed_out.rds", package = "rtrees"))
speed_out = dplyr::mutate(speed_out, n_sp_missing_k = n_sp_missing / 1000, time_s = time/1e9)
# speed_lm = dplyr::group_by(speed_out, expr) |> 
#   dplyr::do(broom::tidy(lm(time_s ~ n_sp_missing_k, data = .))) |> 
#   dplyr::ungroup()
ggplot(speed_out, aes(x = n_sp_missing_k, y = time_s, color = expr, group = expr)) +
  geom_point() +
  geom_smooth(method = "lm") +
  labs(x = "Number of binding species (Thousand)",
       y = "Time (Second)", color = NULL, group = NULL) +
  scale_x_continuous(breaks = c(0.05, seq(0.5, 5, 0.5))) +
  scale_y_continuous(breaks = c(0, 10, seq(25, 200, 25))) +
  geom_text(x = 4, y = 30, label = "rtrees: y = 0.35+ 0.95x", inherit.aes = FALSE, color = "#F8766D") +
  geom_text(x = 1.8, y = 150, label = "V.PhyloMaker2: y = 12.3 + 36.8x", inherit.aes = FALSE, color = "#00BFC4") +
  theme(legend.position = "none") 
```

Based on the figure above, the time used by `rtrees` (and other R packages such as `V.PhyloMaker` and `FishPhyloMaker`) increased linearly with the number of missing species. This is because of the sequential structure of the R `phylo` object, in which all tips and nodes of a phylogeny are labeled with continued integers. When a new node or tip needs to be inserted, all the numbers after the binding location need to be pushed one place after. Therefore, the binding of tips in an R `phylo` object is a sequential process, which cannot be conducted with parallel computing to take advantage of the multiple CPU cores a regular computer has nowadays. It is possible to speed it up by converting some core R code into C++. However, it takes a lot of effort to do this right. 

Another potential resolution is to represent phylogenies in a non-sequential structure unlike `phylo` objects in R. One example is the [`treeman` package](https://github.com/DomBennett/treeman), which developed such a structure to perform tree manipulations. Because of the non-sequential structure (i.e., no continuous labeling of nodes and tips), it is possible to use parallel computing to process phylogenies. However, such a structure has not yet been adapted widely by the R community. Therefore, the infrastructure to support such a non-sequential structure in R is still missing, making it hard to use it here in `rtrees`. 

Despite the limitations described above, `rtrees` is reasonably fast and user-friendly with informative messages. 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind_tip.R
\name{bind_tip_df}
\alias{bind_tip_df}
\title{Bind a tip to a phylogeny (data frame version)}
\usage{
bind_tip_df(
  tree = NULL,
  where,
  tip_label,
  frac = 0.5,
  new_node_above = FALSE,
  node_label = NULL,
  return_tree = TRUE,
  tree_tbl = NULL,
  node_heights = NULL,
  use_castor = FALSE
)
}
\arguments{
\item{tree}{A phylogeny, with class of "phylo".}

\item{where}{Location where to insert the tip. It can be either tip label or node label, but must be characters. If the location does not have a name, assign it first.}

\item{tip_label}{Name of the new tip inserted.}

\item{frac}{The fraction of branch length, must be between 0 and 1. This only applies when location is a tip or \code{new_node_above = TRUE}. The distance from the new inserted node to the location (a node or a tip) is the branch length of the location * (1 - frac).}

\item{new_node_above}{Whether to insert the new node above when the location is a node? Default is \code{FALSE}, which will attach the new tip to the location node.}

\item{node_label}{Name of the new node created. This only applies when location is a tip or \code{new_node_above = TRUE}.}

\item{return_tree}{Whether to return a phylogeny with class "phylo?" Default is \code{TRUE}. Otherwise, it will return a data frame.}

\item{tree_tbl}{A tibble version of the tree, optional.}

\item{node_heights}{A named numeric vector of node hieghts of the tree, generated by \code{\link[ape:branching.times]{ape::branching.times()}}. It is also optional if \code{tree} is specified; but required if \code{tree_tbl} is specified.}

\item{use_castor}{Whether to use package \code{castor} to get the phylogeny at a node; it is faster than \code{tidytree::offspring} to figure out what are the tip offsprings at a node.}
}
\value{
Either a phylogeny or a data frame, which can be then converted to a phylogeny later.
}
\description{
Graft a tip to a phylogeny at location specified.
}
\examples{
tr <- ape::read.tree(text = "((A:1,B:1):1,C:2);")
tr$node.label <- c("root", "N1")
bind_tip_df(tr, where = "N1", tip_label = "D")
tr_tbl <- tidytree::as_tibble(tr)
node_hts <- ape::branching.times(tr)
bind_tip_df(tree_tbl = tr_tbl, where = "N1",
            tip_label = "D", node_heights = node_hts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{add_root_info}
\alias{add_root_info}
\title{Add genus and family basal/root node information to a phylogeny}
\usage{
add_root_info(
  tree,
  classification,
  process_all_tips = TRUE,
  genus_list = NULL,
  family_list = NULL,
  show_warning = FALSE
)
}
\arguments{
\item{tree}{A phylogeny with class "phylo".}

\item{classification}{A data frame of 2 columns: genus, family. It should include
all genus the tips of the tree belong to.}

\item{process_all_tips}{Whether to find basal nodes for all tips? Default is \code{TRUE}.}

\item{genus_list}{An optinoal subset list of genus to find root information.}

\item{family_list}{An optinoal subset list of family to find root information.
This should be for species that do not have co-genus in the tree.}

\item{show_warning}{Whether to print warning information about non-monophyletic clades or not.}
}
\value{
A phylogeny with basal nodes information attached.
}
\description{
Based on the classification of tips, find where is the basal and root node for
each genus and each family. Such information can be later used to graft new
tips onto the phylogeny. This function can be used to process a user provided
tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{sp_list_df}
\alias{sp_list_df}
\title{Convert a vector of species names to a data frame}
\usage{
sp_list_df(sp_list, taxon)
}
\arguments{
\item{sp_list}{A string vector or a data frame with at least one column named "species".}

\item{taxon}{The taxon group of this species list. If not specified, only species and
genus will be returned.}
}
\value{
A data frame with columns: species, genus, and family (if \code{taxon} is specified).
}
\description{
Convert a vector of species names to a data frame
}
\examples{
sp_list_df(
  sp_list = c("Serrasalmus_geryi", "Careproctus_reinhardti", "Gobiomorphus_coxii"),
  taxon = "fish"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{rm_stars}
\alias{rm_stars}
\title{Remove trailing *}
\usage{
rm_stars(tree)
}
\arguments{
\item{tree}{A phylogeny generated by \code{get_tree(..., show_grafted = TRUE)} with trailing stars in tip labels.}
}
\value{
A phylogeny after removing trailing stars.
}
\description{
Remove trailing *
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_one_tree.R
\name{get_one_tree}
\alias{get_one_tree}
\title{Derive a phylogeny from a mega-tree}
\usage{
get_one_tree(
  sp_list,
  tree,
  taxon,
  scenario = c("at_basal_node", "random_below_basal"),
  show_grafted = FALSE,
  tree_by_user = FALSE,
  .progress = "text",
  dt = TRUE
)
}
\arguments{
\item{sp_list}{A character vector or a data frame with at least three columns: species, genus, family. Species column
holds the species for which we want to have a phylogeny. It can also have two optional columns:
close_sp and close_genus. We can specify the closest species/genus of the species based on
expert knowledge. If specified, the new species will be grafted to that particular location.

It can also be a string vector if \code{taxon} is specified. Though it probably is a better idea
to prepare your data frame with \code{\link[=sp_list_df]{sp_list_df()}}. The string vector can also have the same format
as that required by Phylomatic (i.e., family/genus/genus_sp).}

\item{tree}{A mega-tree with class \code{phylo} or a list of mega-trees with class \code{multiPhylo}.
Optional if \code{taxon} is specified, in which case, a default
mega-phylogeny (or a set of 100 randomly selected posterior phylogenies) will be used
(see their own documentations from the \code{megatrees} package).
\itemize{
\item For amphibian, the mega-trees are loaded via \code{\link[megatrees:get_tree_amphibian_n100]{megatrees::get_tree_amphibian_n100()}}.
\item For bee, the mega-tree is \link[megatrees:tree_bee]{megatrees::tree_bee}, with the bootstrap option loaded via \code{\link[megatrees:get_tree_bee_n100]{megatrees::get_tree_bee_n100()}}.
\item For butterfly, the mega-tee is \link[megatrees:tree_butterfly]{megatrees::tree_butterfly}.
\item For bird, the mega-trees are loaded via \code{\link[megatrees:get_tree_bird_n100]{megatrees::get_tree_bird_n100()}}.
\item For fish, the mega-tree is \link[megatrees:tree_fish_12k]{megatrees::tree_fish_12k}, with the all-taxon option loaded via \code{\link[megatrees:get_tree_fish_32k_n50]{megatrees::get_tree_fish_32k_n50()}}.
\item For mammal, the default mega-trees are loaded via \code{\link[megatrees:get_tree_mammal_n100_vertlife]{megatrees::get_tree_mammal_n100_vertlife()}}, with \code{\link[megatrees:get_tree_mammal_n100_phylacine]{megatrees::get_tree_mammal_n100_phylacine()}} be the other option.
\item For plant, the default mega-tree is \link[megatrees:tree_plant_otl]{megatrees::tree_plant_otl} (Smith and Brown 2018);
\link[megatrees:tree_plant_Carruthers]{megatrees::tree_plant_Carruthers} (Carruthers et al. 2026) is also available; 100 posterior trees are loaded via
\code{\link[megatrees:get_tree_plant_n100_Carruthers]{megatrees::get_tree_plant_n100_Carruthers()}}.
\item For reptile, the mega-trees are loaded via \code{\link[megatrees:get_tree_reptile_n100]{megatrees::get_tree_reptile_n100()}}.
\item For shark, ray, and chimaeras, the mega-trees are loaded via \code{\link[megatrees:get_tree_shark_ray_n100]{megatrees::get_tree_shark_ray_n100()}}.
}}

\item{taxon}{The taxon of species in the \code{sp_list}. Currently, can be \code{amphibian}, \code{bird}, \code{fish}, \code{mammal}, \code{plant}, \code{reptile}, or \code{shark_ray}.}

\item{scenario}{How to insert a species into the mega-tree?
\itemize{
\item In both scenarioes, if there is only 1 species in the genus or family, a new node will be inserted to
the middle point of this only species' branch length and the new species will be attached to this new
node.
\item If \code{scenario = "at_basal_node"}, a species is attached to the basal node of the same genus or the same family
if the mega-tree does not have any species of this genus.
\item If \code{scenario = "random_below_basal"}, a species is attached to a randomly selected node that is at or below the
basal node of the same genus of the same family if the mega-tree does not have any species in this genus.
The probability of node been selected is proportional to its branch length.
Because of the random sampling involved, you may want to run several times to get a collection of
derived phylogenies.
}}

\item{show_grafted}{Whether to indicate which species was grafted onto the mega-tree.
If \code{TRUE}, a \code{*} will be appended to the species name on the tip if it was grafted within
the same genus; \verb{**} will be appended if it was grafted within the same family.}

\item{tree_by_user}{Is the mega-tree provided by user? Default is \code{FALSE} but it will be automatically set to \code{TRUE} when the class of \code{tree} is
\code{multiPhylo} since we don't provide any such mega-trees here.}

\item{.progress}{Form of progress bar, default to be text.}

\item{dt}{Whether to use data.table version to bind tips \link{bind_tip}. The default is \code{TRUE} as it maybe slightly faster.}
}
\value{
A phylogeny for the species required, with class \code{phylo}.
}
\description{
For a list of species, generate a phylogeny from a provided mega-tree. If a species is
not in the mega-tree, it will be grafted to the mega-tree with three scenarioes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{taxa_supported}
\alias{taxa_supported}
\title{Taxonomic groups supported}
\format{
An object of class \code{character} of length 9.
}
\usage{
taxa_supported
}
\description{
Supported taxonomic groups with mega-trees provided in the \pkg{megatrees} package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_graft_status}
\alias{get_graft_status}
\title{Extract grafting status information as a data frame}
\usage{
get_graft_status(tree)
}
\arguments{
\item{tree}{A phylogeny generated by \code{get_tree(...)} with trailing stars in tip labels.}
}
\value{
A tibble with three columns: tip_label, species, and status.
}
\description{
Extract grafting status information as a data frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{\%fin\%}
\alias{\%fin\%}
\title{Faster match of vectors}
\description{
See \code{fastmatch::\link[fastmatch:fmatch]{\%fin\%}} for details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtrees-package.R
\docType{package}
\name{rtrees-package}
\alias{rtrees-package}
\alias{rtrees}
\title{Fetching phylogenies from mega-trees}
\description{
Provides tools to derive species-level phylogenies from large synthesis mega-trees for a wide range of taxonomic groups, including plants, birds, mammals, amphibians, reptiles, fish, bees, butterflies, and sharks. When a queried species is absent from the mega-tree, it is grafted onto the tree using one of two placement strategies: attachment at the basal node of the most closely related genus or family ('at_basal_node'), or random attachment below that basal node with probability proportional to branch length ('random_below_basal'). See Li (2023) \doi{10.1111/ecog.06643} for details. Multiple species from a genus not represented in the mega-tree are placed as a polytomy to preserve clade coherence. The package interfaces with the 'megatrees' data package, which bundles or downloads on demand curated mega-trees. Users can also provide their own mega-trees.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://daijiang.github.io/rtrees/}
}

}
\author{
Daijiang Li \href{mailto:daijianglee@gmail.com}{daijianglee@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind_tip.R
\name{bind_tip}
\alias{bind_tip}
\title{Bind a tip to a phylogeny}
\usage{
bind_tip(
  tree = NULL,
  where,
  tip_label,
  frac = 0.5,
  new_node_above = FALSE,
  node_label = NULL,
  return_tree = TRUE,
  tree_tbl = NULL,
  node_heights = NULL,
  use_castor = TRUE,
  sequential = TRUE
)
}
\arguments{
\item{tree}{A phylogeny, with class of "phylo".}

\item{where}{Location where to insert the tip. It can be either tip label or node label, but must be characters. If the location does not have a name, assign it first.}

\item{tip_label}{Name of the new tip inserted.}

\item{frac}{The fraction of branch length, must be between 0 and 1. This only applies when location is a tip or \code{new_node_above = TRUE}. The distance from the new inserted node to the location (a node or a tip) is the branch length of the location * (1 - frac).}

\item{new_node_above}{Whether to insert the new node above when the location is a node? Default is \code{FALSE}, which will attach the new tip to the location node.}

\item{node_label}{Name of the new node created. This only applies when location is a tip or \code{new_node_above = TRUE}.}

\item{return_tree}{Whether to return a phylogeny with class "phylo?" Default is \code{TRUE}. Otherwise, it will return a data frame.}

\item{tree_tbl}{A tibble version of the tree, optional.}

\item{node_heights}{A named numeric vector of node hieghts of the tree, generated by \code{\link[ape:branching.times]{ape::branching.times()}}. It is also optional if \code{tree} is specified; but required if \code{tree_tbl} is specified.}

\item{use_castor}{Whether to use package \code{castor} to get the phylogeny at a node; it is faster than \code{tidytree::offspring} to figure out what are the tip offsprings at a node.}

\item{sequential}{Whether to add the tip with sequential node number in the edge matrix. For example, if we want to bind a tip to a clade and the node number of the tips of this clade is from 101 to 150. We can set the node id of the new tip to 151 and push all the remaining node id to 1 after their current values. This will require us to find out the node ids of all tips that are descents of the node where we want to bind the new tip to, and it can be time costly. Yet I am still not sure whether this is necessary. Normally, the node ids of the \code{phylo} class are sequential. Therefore, the default value here is \code{TRUE}. If set to \code{FALSE}, we can just assign the id of the new tip to Ntip + 1 to save time. In addition, we probably don't need to order the node column of the edge matrix every time.}
}
\value{
Either a phylogeny or a data frame, which can be then converted to a phylogeny later.
}
\description{
Graft a tip to a phylogeny at location specified.
}
\examples{
tr <- ape::read.tree(text = "((A:1,B:1):1,C:2);")
tr$node.label <- c("root", "N1")
bind_tip(tr, where = "N1", tip_label = "D")
tr_tbl <- tidytree::as_tibble(tr)
node_hts <- ape::branching.times(tr)
bind_tip(tree_tbl = tr_tbl, where = "N1",
         tip_label = "D", node_heights = node_hts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{classifications}
\alias{classifications}
\title{Classifications of species}
\format{
A data frame with three columns: genus, family, and taxon (\code{plant}, \code{fish}, \code{bird}, \code{mammal}, \code{amphibian}, \code{reptile}, \code{shark_ray}, \code{bee}, \code{butterfly}).
}
\usage{
classifications
}
\description{
Genus and family information of different groups of taxon.
\itemize{
\item Plant classification information. Its sources include:
+ based on \href{https://github.com/jinyizju/V.PhyloMaker/tree/master/data}{\code{V.PhyloMaker::nodes.info.1}}
+ based on The Plant List
+ \href{https://github.com/traitecoevo/taxonlookup}{taxonlookup}
+ \href{http://plantsoftheworldonline.org/}{Plants of the World online}
\item Fish classification information was based on FishBase. There are 4,825 genus in this file.
\url{https://fishtreeoflife.org/downloads/PFC_taxonomy.csv.xz}
\item Bee classification information was from \href{http://beetreeoflife.org/nomenclature/}{Bee Tree of Life}. Note that we used 'Subfamily' in their nomenclature file as "family" here. If a genus' Subfamily is missing, we used its Family.
\item Bird classification information was based on BirdLife, which resulted in 2,391 genus.
\url{http://datazone.birdlife.org/species/taxonomy} However, based on the
\href{https://data.vertlife.org/birdtree/BLIOCPhyloMasterTax.csv}{taxonomy file}
of the Jetz et al. 2012 phylogeny, there are additional 117 genus that are not in the file
of BirdLife. Both are combined here, which leads to 2,508 genus.
\item Mammal classification information was based on PHYLACINE, which has 1,400 genus.
\url{https://github.com/MegaPast2Future/PHYLACINE_1.2/blob/master/Data/Taxonomy/Synonymy_table_valid_species_only.csv} Additional
genus from Vertlife were added too. For the same genus from both PHYLACINE and Vertlife that have different family information,
I used the family from Vertlife as I found that they are mostly more accurate.
\item Amphibian classification information was from \href{https://data.vertlife.org/amphibiantree/download/amph_shl_new_Classification.csv}{VertLife}.
\item Reptile classification information was largely from \href{https://en.wikipedia.org/wiki/List_of_reptile_genera}{wikipedia}.
\item Shark and Ray classification information was largely from NCBI.
\item Butterfly classification information was from Kawahara et al. 2023, using the tip labels of their phylogeny.
}
}
\keyword{datasets}
